# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from gillespy2.core.jsonify import Jsonify
from gillespy2.core.sortableobject import SortableObject

from gillespy2.core.gillespyError import SpeciesError

class Species(SortableObject, Jsonify):
    """
    Chemical species. Can be added to Model object to interact with other
    species or time.

    :param name: The name by which this species will be called in reactions and within the model.
    :type name: str

    :param initial_value: Initial population (discrete) or concentration (continuous) of this species.
    :type initial_value: int | float

    :param constant: If true, the value of the species cannot be changed (currently TauHybridSolver only)
    :type constant: bool

    :param boundary_condition: If true, species can be changed by events and rate rules, but not by reactions.
        (TauHybridSolver only)
    :type boundary_condition: bool

    :param mode: ***FOR USE WITH TauHybridSolver ONLY***
        Sets the mode of representation of this species for the TauHybridSolver,
        can be discrete, continuous, or dynamic.
        mode='dynamic' - Allows a species to be represented as either discrete or continuous
        mode='continuous' - Species will only be represented as continuous
        mode='discrete' - Species will only be represented as discrete
    :type mode: str

    :param allow_negative_populations: If true, population can be reduced below 0.
    :type allow_negative_populations: bool

    :param switch_tol: ***FOR USE WITH TauHybridSolver ONLY***
        Tolerance level for considering a dynamic species deterministically, value is compared to an estimated sd/mean
        population of a species after a given time step. This value will be used if a switch_min is not provided.
        The default value is 0.03
    :type switch_tol: float

    :param switch_min:  ***FOR USE WITH TauHybridSolver ONLY***
        Minimum population value at which species will be represented as continuous.
        If a value is given, switch_min will be used instead of switch_tol
    :type switch_min: float

    :raises SpeciesError: Arg is of invalid type.  Required arg set to None.  Arg value is outside of accepted bounds.
    """

    def __init__(self, name=None, initial_value=0, constant=False, boundary_condition=False, mode=None,
                 allow_negative_populations=False, switch_min=0, switch_tol=0.03):
        # A species has a name (string) and an initial value (positive integer)
        self.name = name
        self.constant = constant
        self.boundary_condition = boundary_condition
        self.mode = mode
        self.allow_negative_populations = allow_negative_populations
        self.switch_min = switch_min
        self.switch_tol = switch_tol

        if initial_value is None:
            raise SpeciesError("initial_value can't be None type.")
        if isinstance(initial_value, str):
            try:
                initial_value = float(initial_value)
            except ValueError:
                pass
        self.validate(initial_value=initial_value)

        self.initial_value = float(initial_value) if self.mode == "continuous" else int(initial_value)

    def __str__(self):
        print_string = self.name
        print_string += ': ' + str(self.initial_value)
        return print_string

    def set_initial_value(self, initial_value):
        """
        Setter method for initial_value of a population

        :param initial_value: Initial population (discrete) or concentration (continuous) of this species.
        :type initial_value: int | float

        :raises SpeciesError: initial_value is of invalid type.  initial_value set to None.  \
                              initial_value is a float when mode != 'continuous'. \
                              initial_value is negative when allow_negative_populations=False.
        """
        if initial_value is None:
            raise SpeciesError("initial_value can't be None type.")
        if isinstance(initial_value, str):
            try:
                initial_value = float(initial_value)
            except ValueError:
                pass
        self.validate(initial_value=initial_value, coverage="initial_value")

        self.initial_value = float(initial_value) if self.mode == "continuous" else int(initial_value)

    def validate(self, initial_value=None, coverage="all"):
        """
        Validate the species.

        :param initial_value: Initial population (discrete) or concentration (continuous) of this species.
        :type initial_value: int | float

        :param coverage: The scope of attributes to validate.  Set to an attribute name to restrict validation \
                         to a specific attribute.
        :type coverage: str

        :raises SpeciesError: Attribute is of invalid type.  Required attribute set to None.  \
                              Attribute is value outside of accepted bounds.
        """
        # Check name
        if coverage in ("all", "name"):
            if self.name is None:
                raise SpeciesError("name can't be None type.")
            if not isinstance(self.name, str):
                raise SpeciesError(f"name must be of type str not {type(self.name)}.")
            if self.name == "":
                raise SpeciesError("name can't be an empty string.")

        # Check initial_value
        if coverage in ("all", "initial_value"):
            if initial_value is None:
                initial_value = self.initial_value

            if initial_value is None:
                raise SpeciesError("initial_value can't be None type.")
            if not isinstance(initial_value, (float, int)):
                raise SpeciesError(f"initial_value must be of type float or int not {type(initial_value)}.")
            if self.mode != "continuous" and int(initial_value) != initial_value:
                raise SpeciesError(
                    """
                    initial_value with mode='discrete' must be an integer value.
                    Change to mode='continuous' to use floating point values.
                    """
                )
            if not self.allow_negative_populations and initial_value < 0:
                raise SpeciesError(
                    'A species initial value must be non-negative unless allow_negative_populations=True'
                )

        # Check constant
        if coverage in ("all", "constant"):
            if not isinstance(self.constant, bool):
                errmsg = f"constant must be of type bool not {type(self.constant)}."
                raise SpeciesError(errmsg)

        # Check boundary_condition
        if coverage in ("all", "boundary_condition"):
            if not isinstance(self.boundary_condition, bool):
                errmsg = f"boundary_condition must be of type bool not {type(self.boundary_condition)}."
                raise SpeciesError(errmsg)

        # Check mode
        if coverage in ("all", "mode"):
            mode_list = ['continuous', 'dynamic', 'discrete', None]

            if self.mode not in mode_list:
                raise SpeciesError(
                    f"""
                    mode must be 'continuous', 'dynamic', 'discrete', or
                    unspecified (defaults to 'dynamic' for TauHybridSolver) not {self.mode}.
                    """
                )

        # Check allow_negative_populations
        if coverage in ("all", "allow_negative_populations"):
            if not isinstance(self.allow_negative_populations, bool):
                errmsg = f"allow_negative_populations must be of type bool not {type(self.allow_negative_populations)}."
                raise SpeciesError(errmsg)

        # Check switch_tol
        if coverage in ("all", "switch_tol"):
            if self.switch_tol is None:
                raise SpeciesError("switch_tol can't be None type.")
            if not isinstance(self.switch_tol, (float, int)):
                raise SpeciesError(f"switch_tol must of type float or int not {type(self.switch_tol)}")
            if self.switch_tol < 0:
                raise SpeciesError(f"switch_tol must be a positive value not {self.switch_tol}")

        # Check switch_min
        if coverage in ("all", "switch_min"):
            if self.switch_min is None:
                raise SpeciesError("switch_min can't be None type.")
            if not isinstance(self.switch_min, (float, int)):
                raise SpeciesError(f"switch_min must of type float or int not {type(self.switch_min)}")
            if self.switch_min < 0:
                raise SpeciesError(f"switch_min must be a positive value not {self.switch_min}")



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy
import gillespy2

class TestNegativeState(unittest.TestCase):
    def setUp(self):

        def create_hybrid_model(include_reaction=False):
            model = gillespy2.Model(name="test_negative_state")
            # Parameters
            k1 = gillespy2.Parameter(name="k1", expression="1e-05")
            model.add_parameter([k1])
            # Variables
            A = gillespy2.Species(name="A", initial_value=10)
            model.add_species([A])
            # Reactions
            if include_reaction:
                model.add_reaction(gillespy2.Reaction(name="r1", reactants={}, 
                    products={'A': 1}, rate="k1"))
            # Rate Rule
            model.add_rate_rule(gillespy2.RateRule(name="rr1", formula="-1", 
                variable="A"))
            # Timespan
            model.timespan(gillespy2.TimeSpan.arange(1,t=20))
            return model

        self.rr_only_model = create_hybrid_model(include_reaction=False)
        self.rr_rxn_model = create_hybrid_model(include_reaction=True)

    def test_hybrid_python_rr_only(self):
        sol = gillespy2.TauHybridSolver(model=self.rr_only_model)
        result = sol.run()
        self.assertLess(result['A'][-1], -9.9)

    def test_hybrid_python_rr_rxn(self):
        sol = gillespy2.TauHybridSolver(model=self.rr_rxn_model)
        with self.assertRaises(gillespy2.SimulationError) as ex:
            result = sol.run()
        self.assertIn('Negative State detected at begining of step. Species involved in reactions can not be negative.', str(ex.exception))


    def test_hybrid_C_rr_only(self):
        sol = gillespy2.TauHybridCSolver(model=self.rr_only_model)
        result = sol.run()
        self.assertLess(result['A'][-1], -9.9)

    def test_hybrid_C_rr_rxn(self):
        sol = gillespy2.TauHybridCSolver(model=self.rr_rxn_model)
        with self.assertRaises(gillespy2.SimulationError) as ex:
            result = sol.run()
        self.assertEqual('Negative State detected at beginning of step. Species involved in reactions can not be negative.', str(ex.exception))


if __name__ == '__main__':
    unittest.main()




# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import unittest
import gillespy2
from gillespy2 import TauHybridCSolver
import numpy as np

class EventFeatures(unittest.TestCase):
    def create_base_event_model(s1, s2, rate):
        model = gillespy2.Model(name="BasicEventModel")

        s1 = gillespy2.Species(name="S1", initial_value=s1, mode="continuous")
        s2 = gillespy2.Species(name="S2", initial_value=s2, mode="continuous")
        model.add_species([s1, s2])
        
        rate = gillespy2.Parameter(name="k1", expression=rate)
        model.add_parameter(rate)

        r1 = gillespy2.Reaction(
            name="r1", reactants={s1: 1}, products={s2: 1}, rate=rate,
        )
        model.add_reaction(r1)
        return model

    def test_event_with_time_trigger(self):
        model = EventFeatures.create_base_event_model(s1=0, s2=0, rate=0.0)
        event = gillespy2.Event(name="ev1", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="100.0"),
            gillespy2.EventAssignment(variable=model.get_parameter('k1'), expression="1.0")
        ], trigger=gillespy2.EventTrigger(expression="t>5"))
        model.add_event(event)

        solver = TauHybridCSolver(model=model)
        result = model.run(solver=solver)[0]
        s1, s2 = result["S1"][-1], result["S2"][-1]

        self.assertGreater(s2, s1, "Expected S2 > S1")
        self.assertGreater(s1, 0.0, "Expected S1 > 0")
        self.assertAlmostEqual(s1 + s2, 100.0, places=1)

    def test_event_with_species_trigger(self):
        model = EventFeatures.create_base_event_model(s1=100, s2=0, rate=10.0)
        event = gillespy2.Event(name="ev1", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="100.0"),
            gillespy2.EventAssignment(variable=model.get_parameter('k1'), expression="0.0")
        ], trigger=gillespy2.EventTrigger(expression="S1<90"))
        model.add_event(event)

        solver = TauHybridCSolver(model=model)
        result = model.run(solver=solver)[0]
        s1, s2 = result["S1"][-1], result["S2"][-1]

        self.assertEqual(s1, 100, "Expected S1 == 100 (trigger set S1 to 100 and rate to 0")
        self.assertGreater(s2, 0, "Expected S2 > 0")
        self.assertFalse(np.any(result["S1"] <= 90.0), "Expected S1 > 90 for entire simulation")

    def test_delay_trigger_persistent(self):
        model = EventFeatures.create_base_event_model(s1=100, s2=0, rate=1.0)
        event1 = gillespy2.Event(name="ev1", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="0"),
            gillespy2.EventAssignment(variable=model.get_species('S2'), expression="0"),
            gillespy2.EventAssignment(variable=model.get_parameter('k1'), expression="0.0")
        ], trigger=gillespy2.EventTrigger(expression="S1<60 and S2<S1", persistent=False), delay="t+1.0")
        event2 = gillespy2.Event(name="ev2", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="200"),
            gillespy2.EventAssignment(variable=model.get_species('S2'), expression="200"),
            gillespy2.EventAssignment(variable=model.get_parameter('k1'), expression="0.0"),
        ], trigger=gillespy2.EventTrigger(expression="S2>90 and t<3.5", persistent=True), delay="1.0")
        model.add_event([event1, event2])

        solver = TauHybridCSolver(model=model)
        result = model.run(solver=solver)[0]
        s1, s2 = result["S1"][-1], result["S2"][-1]

        # If delay is working correctly:
        # * event1 is never triggered. event1 sets everything to 0.
        #   If event1 fires, event2 can never fire.
        # * event2 is triggered, setting everything to 100 (and rate to 0).
        self.assertNotIn(0, [s1, s2], "Non-persistent event fired unexpectedly")
        self.assertEqual(s1, 200, "Persistent event failed to fire")
        self.assertEqual(s2, 200, "Persistent event failed to fire")

    def test_trigger_priorities(self):
        model = EventFeatures.create_base_event_model(s1=100, s2=0, rate=1.0)
        event1 = gillespy2.Event(name="ev1", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="100"),
            gillespy2.EventAssignment(variable=model.get_species('S2'), expression="100"),
        ], trigger=gillespy2.EventTrigger(expression="S1 < 50"), priority="2*t*S1")
        event2 = gillespy2.Event(name="ev2", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="0"),
        ], trigger=gillespy2.EventTrigger(expression="S1 < 50"), priority="t*S1")
        model.add_event([event1, event2])

        solver = TauHybridCSolver(model=model)
        result = model.run(solver=solver)[0]
        s1, s2 = result["S1"][-1], result["S2"][-1]

        # If priority is working correctly, event2 should ALWAYS fire before event1.
        # Proper result is S1 = 0, S2 = 100, so no further reactions are possible.
        self.assertEqual(s1, 0,   "Events fired in an incorrect order")
        self.assertEqual(s2, 100, "Events fired in an incorrect order")

    def test_use_values_from_trigger_time(self):
        model = EventFeatures.create_base_event_model(s1=100, s2=0, rate=1.0)
        event = gillespy2.Event(name="ev1", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="S2"),
            gillespy2.EventAssignment(variable=model.get_parameter('k1'), expression="0.0"),
        ], trigger=gillespy2.EventTrigger(expression="S1 < 60"), delay="1.5", use_values_from_trigger_time=True)
        model.add_event(event)

        solver = TauHybridCSolver(model=model)
        result = model.run(solver=solver)[0]
        s1, s2 = result["S1"][-1], result["S2"][-1]

        self.assertGreater(s2, s1, "Event assignment did not assign values from trigger time")

    def test_initial_values(self):
        model = EventFeatures.create_base_event_model(s1=0, s2=100.0, rate=1.0)

        event = gillespy2.Event(name="ev1", assignments=[
            gillespy2.EventAssignment(variable=model.get_species('S1'), expression="S2/2"),
        ], trigger=gillespy2.EventTrigger(expression="S1==0", initial_value=False))
        model.add_event(event)

        solver = TauHybridCSolver(model=model)
        result = model.run(solver=solver)

        s1, s2 = result["S1"][-1], result["S2"][-1]
        self.assertAlmostEqual(s1 + s2, 150.0, places=1, msg="Event assignment assigned incorrect value")
        self.assertGreater(s2, 100, "Event with initial condition did not fire")
        self.assertEqual(result["S1"][0], 50, "Event assignment with initial condition failed to fire at t=0")



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from datetime import time
import unittest
import numpy as np
import io
import gillespy2
from gillespy2 import Model, Species, Parameter, Reaction, Event, \
                      EventTrigger, EventAssignment, RateRule, \
                      AssignmentRule, FunctionDefinition, TimeSpan
from gillespy2.solvers import TauHybridCSolver, TauHybridSolver


class TestHybridEventRound(unittest.TestCase):

    def create_model(self, discrete=True):
        model = Model(name="test_issue710_round_event_assign")

        # Parameters
        k1 = Parameter(name="k1", expression="1e-05")
        model.add_parameter([k1])

        # Variables
        if discrete:
            A = Species(name="A", initial_value=5, mode='discrete')
        else:
            A = Species(name="A", initial_value=5, mode="continuous")
        model.add_species([A])

        # Reactions
        r1 = Reaction(name="r1", reactants={}, products={'A': 1}, rate="k1")
        model.add_reaction([r1])

        # Event Triggers
        e1_trig = EventTrigger(expression="t>10", initial_value=False, persistent=False)

        # Event Assignments
        e1_assign_1 = EventAssignment(variable="A", expression="A + 0.5")

        # Events
        model.add_event(Event(name="e1", trigger=e1_trig, assignments=[e1_assign_1], delay=None, priority="0", use_values_from_trigger_time=False))

        # Timespan
        tspan = TimeSpan(np.arange(0, 20.05, 0.05))
        model.timespan(tspan)
        return model

    def create_dynamic_model(self):
        model = Model(name="test_issue710_round_event_assign_switch")

        # Parameters
        k1 = Parameter(name="k1", expression="1e-05")
        model.add_parameter([k1])

        # Variables
        A = Species(name="A", initial_value=100, mode='dynamic', 
                    switch_min=50)
        model.add_species([A])

        # Reactions
        r1 = Reaction(name="r1", reactants={}, products={'A': 1}, rate="k1")
        model.add_reaction([r1])

        # Event Triggers
        t5_trig = EventTrigger(expression="t>5", initial_value=False, persistent=False)
        t10_trig = EventTrigger(expression="t>10", initial_value=False, persistent=False)
        t15_trig = EventTrigger(expression="t>15", initial_value=False, persistent=False)

        # Event Assignments
        eassign_1 = EventAssignment(variable="A", expression="A + 0.33")
        eassign_2 = EventAssignment(variable="A", expression="A - 90")

        # Events
        model.add_event(Event(name="e1", trigger=t5_trig, 
                             assignments=[eassign_1], 
                             delay=None, priority="0", 
                             use_values_from_trigger_time=False))
        model.add_event(Event(name="e2", trigger=t10_trig, 
                             assignments=[eassign_2], 
                             delay=None, priority="0", 
                             use_values_from_trigger_time=False))
        model.add_event(Event(name="e3", trigger=t15_trig, 
                             assignments=[eassign_1], 
                             delay=None, priority="0", 
                             use_values_from_trigger_time=False))

        # Timespan
        tspan = TimeSpan(np.arange(0, 20.05, 0.05))
        model.timespan(tspan)
        return model

    def setUp(self):
        self.d_model = self.create_model(discrete=True)
        self.c_model = self.create_model(discrete=False)

        self.dyn_model  = self.create_dynamic_model()

    def test_dynamic_event_rounding(self):
        expected_times_values = {
            99:  100.0,
            102: 100.33,
            202: 10.0,
            302: 10.0,
        }
        number_of_trajectories = 3

        for sname, sclass in {'TauHybridCSolver':TauHybridCSolver,'TauHybridSolver':TauHybridSolver}.items():
            with self.subTest(f"Checking event assignment rounding for dynamic species with {sname} solver."):
                solver = sclass(model=self.dyn_model)
                results = solver.run(number_of_trajectories=number_of_trajectories)
                for result in results:
                    for time_value, expected_value in expected_times_values.items():
                        self.assertAlmostEqual(result['A'][time_value], expected_value, places=2, msg=f"Simulation output at {time_value} is {result['A'][time_value]}, should be {expected_value}")


    def test_event_rounding(self):
        number_of_trajectories = 3
        for model, expected_last in {self.d_model: 6, self.c_model: 5.5}.items():
            expected_init = model.get_species("A").initial_value

            for sname, sclass in {'TauHybridCSolver':TauHybridCSolver,'TauHybridSolver':TauHybridSolver}.items():
                with self.subTest(f"Checking event assignment rounding for {model.listOfSpecies['A'].mode} species with {sname} solver."):
                    solver = sclass(model=model)
                    results = solver.run(number_of_trajectories=number_of_trajectories)
                    for result in results:
                        first_value = result["A"][0]
                        last_value =  result["A"][-1], 
                        self.assertAlmostEqual(first_value, expected_init, places=3, msg=f"Simulation output begins with {first_value}, should be {expected_init}")
                        self.assertAlmostEqual(result["A"][-1], expected_last, places=3, msg=f"Simulation output ends with {last_value}, should be {expected_last}")




# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from collections.abc import Iterator

import numpy as np

from gillespy2.core.jsonify import Jsonify
from .gillespyError import TimespanError

class TimeSpan(Iterator, Jsonify):
    """
    Model timespan that describes the duration to run the simulation and at which timepoint to sample
    the species populations during the simulation.

    :param items: Evenly-spaced list of times at which to sample the species populations during the simulation.
            Best to use the form np.linspace(<start time>, <end time>, <number of time-points, inclusive>)
    :type items: list, tuple, range, or numpy.ndarray

    :raises TimespanError: items is an invalid type.
    """
    def __init__(self, items):
        if isinstance(items, np.ndarray):
            self.items = items
        elif isinstance(items, (list, tuple, range)):
            self.items = np.array(items)
        else:
            raise TimespanError("Timespan must be of type: list, tuple, range, or numpy.ndarray.")

        self.validate()

    def __str__(self):
        return self.items.__str__()

    def __eq__(self, o):
        return self.items.__eq__(o).all()

    def __getitem__(self, key):
        return self.items.__getitem__(key)

    def __iter__(self):
        return self.items.__iter__()

    def __len__(self):
        return self.items.__len__()

    def __next__(self):
        return self.items.__next__()

    @classmethod
    def linspace(cls, t=20, num_points=None):
        """
        Creates a timespan using the form np.linspace(0, <t>, <num_points, inclusive>).

        :param t: End time for the simulation.
        :type t: float | int

        :param num_points: Number of sample points for the species populations during the simulation.
        :type num_points: int

        :returns: Timespan for the model.
        :rtype: gillespy2.TimeSpan

        :raises TimespanError: t or num_points are None, <= 0, or invalid type.
        """
        if t is None or not isinstance(t, (int, float)) or t <= 0:
            raise TimespanError("t must be a positive float or int.")
        if num_points is not None and (not isinstance(num_points, int) or num_points <= 0):
            raise TimespanError("num_points must be a positive int.")

        if num_points is None:
            num_points = int(t / 0.05) + 1
        items = np.linspace(0, t, num_points)
        return cls(items)

    @classmethod
    def arange(cls, increment, t=20):
        """
        Creates a timespan using the form np.arange(0, <t, inclusive>, <increment>).

        :param increment: Distance between sample points for the species populations during the simulation.
        :type increment: float | int

        :param t: End time for the simulation.
        :type t: float | int

        :returns: Timespan for the model.
        :rtype: gillespy2.TimeSpan

        :raises TimespanError: t or increment are None, <= 0, or invalid type.
        """
        if t is None or not isinstance(t, (int, float)) or t <= 0:
            raise TimespanError("t must be a positive floar or int.")
        if not isinstance(increment, (float, int)) or increment <= 0:
            raise TimespanError("increment must be a positive float or int.")

        items = np.arange(0, t + increment, increment)
        return cls(items)

    def validate(self):
        """
        Validate the models time span

        :raises TimespanError: Timespan is an invalid type, empty, not uniform, contains a single \
                               repeated value, or contains a negative initial time.
        """
        if not isinstance(self.items, np.ndarray):
            if not isinstance(self.items, (list, tuple, range)):
                raise TimespanError("Timespan must be of type: list, tuple, range, or numpy.ndarray.")
            self.items = np.array(self.items)

        if len(self.items) == 0:
            raise TimespanError("Timespans must contain values.")
        if self.items[0] < 0:
            raise TimespanError("Simulation must run from t=0 to end time (t must always be positive).")

        first_diff = self.items[1] - self.items[0]
        other_diff = self.items[2:] - self.items[1:-1]
        isuniform = np.isclose(other_diff, first_diff).all()

        if not isuniform:
            raise TimespanError("GillesPy2 only supports uniform timespans.")
        if first_diff == 0 or np.count_nonzero(other_diff) != len(other_diff):
            raise TimespanError("Timespan can't contain a single repeating value.")



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import unittest
from unittest import TestCase

from profiling import python_profiler

from example_models import create_oregonator
from example_models import create_vilar_oscillator
from example_models import create_tyson_2_state_oscillator

from gillespy2.solvers.numpy.ode_solver import ODESolver
from gillespy2.solvers.numpy.ssa_solver import NumPySSASolver
from gillespy2.solvers.numpy.tau_leaping_solver import TauLeapingSolver


class TestPythonSolverPerf(TestCase):
    def setUp(self) -> None:
        self.solvers = {
            NumPySSASolver: [
                create_tyson_2_state_oscillator()
            ],
            ODESolver: [
                create_oregonator()
            ],
            TauLeapingSolver: [
                create_vilar_oscillator()
            ],
        }

    def test_python_solver_perf(self):
        for solver, models in self.solvers.items():
            for model in models:
                with self.subTest(model=model.name, solver=solver.name):
                    perf_data = python_profiler.run_profiler(model, solver)
                    print(perf_data)

if __name__ == "__main__":
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest
import numpy as np
import gillespy2
from gillespy2.core.gillespyError import *
from example_models import create_decay, create_decay_no_tspan, create_multi_firing_event
from gillespy2 import TauHybridSolver


class TestBasicTauHybridSolver(unittest.TestCase):

    def test_null_timeout(self):
        model = create_decay()
        species = gillespy2.Species('test_species', initial_value=1)
        model.add_species([species])
        solver = TauHybridSolver(model=model)
        results = solver.run()
        self.assertTrue(len(results) > 0)

    def test_add_rate_rule(self):
        model = create_decay()
        species = gillespy2.Species('test_species', initial_value=1)
        rule = gillespy2.RateRule(name='rr1',formula='test_species+1',variable='test_species')
        model.add_species([species])
        model.add_rate_rule([rule])
        results = model.run()
        valid_solvers = ('TauHybridSolver', 'TauHybridCSolver')
        self.assertIn(results[0].solver_name, valid_solvers)

    def test_add_rate_rule_dict(self):
        model = create_decay()
        species2 = gillespy2.Species('test_species2', initial_value=2, mode='continuous')
        species3 = gillespy2.Species('test_species3', initial_value=3, mode='continuous')
        rule2 = gillespy2.RateRule('rule2', species2, 'cos(t)')
        rule3 = gillespy2.RateRule(name='rule3', variable=species3, formula='sin(t)')
        rate_rule_dict = {'rule2': rule2, 'rule3': rule3}
        model.add_species([species2, species3])
        with self.assertRaises(ModelError):
            model.add_rate_rule(rate_rule_dict)

    def test_add_bad_species_rate_rule_dict(self):
        model = create_decay()
        rule = gillespy2.RateRule(formula='sin(t)')
        with self.assertRaises(ModelError):
            model.add_rate_rule(rule)

    def test_add_assignment_rule(self):
        model = create_decay()
        species = gillespy2.Species('test_species4', initial_value=1)
        rule = gillespy2.AssignmentRule(name='ar1', variable=species.name, formula='2')
        model.add_species([species])
        model.add_assignment_rule([rule])
        results = model.run()
        self.assertEqual(results[species.name][0], 2) 
        self.assertEqual(results[species.name][-1], 2)
        self.assertEqual(results[0].solver_name,'TauHybridSolver')

    def test_add_function_definition(self):
        model = create_decay()
        funcdef = gillespy2.FunctionDefinition(name='fun', function='Sp+1')
        model.add_function_definition(funcdef)
        results = model.run()
        self.assertEqual(results[0].solver_name,'TauHybridSolver')

    def test_add_continuous_species_dependent_event(self):
        model = create_decay()
        model.listOfSpecies['Sp'].mode = 'continuous'
        eventTrig = gillespy2.EventTrigger(expression='Sp <= 90', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        valid_solvers = ('TauHybridSolver', 'TauHybridCSolver')
        self.assertIn(results[0].solver_name, valid_solvers)
        self.assertEqual(results['Sp'][-1], 1000)

    def test_add_stochastic_species_dependent_event(self):
        model = create_decay()
        model.listOfSpecies['Sp'].mode = 'discrete'
        eventTrig = gillespy2.EventTrigger(expression='Sp <= 90', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results['Sp'][-1], 1000)
        
    def test_add_continuous_time_dependent_event(self):
        model = create_decay()
        model.listOfSpecies['Sp'].mode = 'continuous'
        eventTrig = gillespy2.EventTrigger(expression='t >= 10', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results['Sp'][-1], 1000)
        
    def test_add_stochastic_time_dependent_event(self):
        model = create_decay()
        model.listOfSpecies['Sp'].mode = 'discrete'
        eventTrig = gillespy2.EventTrigger(expression='t >= 10', initial_value=True, )
        event1 = gillespy2.Event(name='event1', trigger=eventTrig)
        ea1 = gillespy2.EventAssignment(variable='Sp', expression='1000')
        ea2 = gillespy2.EventAssignment(variable='k1', expression='0')
        event1.add_assignment([ea1, ea2])
        model.add_event(event1)
        results = model.run()
        self.assertEqual(results['Sp'][-1], 1000)
        
    def test_add_param_event(self):
        def create_event_test_model(parameter_values=None):
            model = gillespy2.Model(name='Event Test Model')
            model.add_species([gillespy2.Species(name='S', initial_value=0)])
            model.add_parameter(gillespy2.Parameter(name='event_tracker',
                                                    expression=99))
            model.add_parameter(gillespy2.Parameter(name='event_tracker2',
                                                    expression=0))
            model.add_reaction(gillespy2.Reaction(name='r1', products={'S':1},
                                                rate=model.listOfParameters['event_tracker2']))
            eventTrig1 = gillespy2.EventTrigger(expression='t>=2')
            event1 = gillespy2.Event(name='event1', trigger=eventTrig1)
            event1.add_assignment(gillespy2.EventAssignment(
                                    variable='event_tracker', expression='t'))
            eventTrig2 = gillespy2.EventTrigger(expression='t >= event_tracker + 2')
            event2 = gillespy2.Event(name='event2', trigger=eventTrig2)
            event2.add_assignment(gillespy2.EventAssignment(
                                    variable='event_tracker2', expression='t'))
            model.add_event([event1, event2])
            return model

        model = create_event_test_model()
        results = model.run(increment=0.05, t=20)
        self.assertGreater(results['S'][-1], 0)

    def test_math_name_overlap(self):
        model = create_decay()
        gamma = gillespy2.Species('gamma',initial_value=2, mode='continuous')
        model.add_species([gamma])
        k2 = gillespy2.Parameter(name='k2', expression=1)
        model.add_parameter([k2])
        gamma_react = gillespy2.Reaction(name='gamma_react', reactants={'gamma': 1}, products={}, rate=k2)
        model.add_reaction([gamma_react])
        solver = TauHybridSolver(model=model)
        model.run(solver=solver)

    def test_add_bad_expression_rate_rule_dict(self):
        model = create_decay()
        species2 = gillespy2.Species('test_species2', initial_value=2, mode='continuous')
        rule = gillespy2.RateRule(variable=species2, formula='')
        with self.assertRaises(ModelError):
            model.add_rate_rule(rule)

    def test_ensure_hybrid_dynamic_species(self):
        model = create_decay()
        species1 = gillespy2.Species('test_species1', initial_value=1,mode='dynamic')
        model.add_species(species1)
        results = model.run()
        valid_solvers = ('TauHybridSolver', 'TauHybridCSolver')
        self.assertIn(results[0].solver_name, valid_solvers)

    def test_ensure_hybrid_continuous_species(self):
        model = create_decay()
        species1 = gillespy2.Species('test_species1', initial_value=1,mode='continuous')
        model.add_species(species1)
        results = model.run()
        valid_solvers = ('TauHybridSolver', 'TauHybridCSolver')
        self.assertIn(results[0].solver_name, valid_solvers)

    def test_ensure_continuous_dynamic_timeout_warning(self):
        model = create_decay()
        species1 = gillespy2.Species('test_species1', initial_value=1, mode='dynamic')
        model.add_species(species1)
        with self.assertLogs('GillesPy2', level='WARN'):
            solver = TauHybridSolver(model=model)
            results = model.run(solver=solver, timeout=1)

    def test_run_example__with_increment_only(self):
        model = create_decay_no_tspan()
        solver = TauHybridSolver(model=model)
        results = solver.run(increment=0.2, t=20)

    def test_run_example__with_tspan_only(self):
        model = create_decay()
        solver = TauHybridSolver(model=model)
        results = solver.run()

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            model = create_decay()
            solver = TauHybridSolver(model=model)
            results = solver.run(increment=0.2)


class TestAllHybridSolvers(unittest.TestCase):
    from gillespy2.solvers import TauHybridCSolver

    solvers = [
        TauHybridSolver,
        TauHybridCSolver,
    ]

    def test_continuous_state_values(self):
        """
        Continuous values should be evaluate appropriately, without being truncated/casted to an integer.
        """
        
        def create_truncated_state_model(parameter_values=None):
            model = gillespy2.Model(name="TruncatedStateModel")
            S1 = gillespy2.Species(name="S1", initial_value=0, mode="discrete")
            rate = gillespy2.Species(name="rate", initial_value=0.9999, mode="continuous")
            model.add_species([S1, rate])
            model.add_rate_rule(gillespy2.RateRule(variable="rate", formula="-1/((t+0.9999)**2)"))
            model.add_reaction(
                # Because S1 is a "discrete" species, our reaction will be marked "stochastic."
                gillespy2.Reaction(products={S1: 1}, propensity_function="10.0*rate")
            )
            return model

        model = create_truncated_state_model()
        for solver in self.solvers:
            with self.subTest(solver=solver.name):
                solver = solver(model=model)
                result = model.run(solver=solver, seed=1, increment=0.05, t=20)
                self.assertGreater(result["S1"][-1], 0.0,
                                   "Reaction never fired; indicates that continuous species is being truncated")

    def test_multi_firing_event(self):
        model = create_multi_firing_event()
        for solver in self.solvers:
            with self.subTest(solver=solver.name):
                solver = solver(model=model)
                res = model.run(solver=solver, seed=1)
                self.assertNotEqual(res['Sp'][45], 0)
                self.assertEqual(res['Sp'][75], 0)
                self.assertNotEqual(res['Sp'][96], 0)
                self.assertEqual(res['Sp'][120], 0)
                self.assertNotEqual(res['Sp'][144], 0)
                self.assertEqual(res['Sp'][165], 0)

if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from datetime import time
import unittest
import numpy as np
import io
import gillespy2
from gillespy2.solvers import TauHybridCSolver, TauHybridSolver
from example_models import create_opioid

class TestHbridSolverOpioidModel(unittest.TestCase):

    def test_run_output(self):
        model = create_opioid()
        for solver in [TauHybridSolver, TauHybridCSolver]:
            results = model.run(solver=solver, number_of_trajectories=3, seed=1024)
            for result in results:
                with self.subTest("Processing simulation output for solver {solver.name} for trajectory={n}"):
                    min_s = min(result['Susceptibles'])
                    self.assertGreater(min_s, 500)
                    max_p = max(result['Prescribed_Users'])
                    self.assertLess(max_p, 500)




# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from datetime import time
import unittest
import numpy as np
import io
import gillespy2
from gillespy2.solvers import ODECSolver, SSACSolver, TauLeapingCSolver, TauHybridCSolver
from example_models import create_degradation
from gillespy2.solvers.cpp.c_decoder import BasicSimDecoder, IterativeSimDecoder

class TestCDecode(unittest.TestCase):
    decoders = [BasicSimDecoder, IterativeSimDecoder]

    def setUp(self):
        self.model = create_degradation()
        tspan = gillespy2.TimeSpan.linspace(t=300, num_points=301)
        self.model.timespan(tspan)
        self.solvers = [
            # Solver and List of trajectory numbers to try
            (ODECSolver(model=self.model), [1]),
            (SSACSolver(model=self.model), [1, 3]),
            (TauLeapingCSolver(model=self.model), [1, 3]),
            (TauHybridCSolver(model=self.model), [1, 3]),
        ]


    def test_c_decoder(self):
        """
        Ensures that, given a certain input from stdout, the results will be
            properly formatted by the sim decoders.
        """
        example_input = """
        0,1.0,2.0,3.0,1,2.0,4.0,6.0,2,3.0,6.0,9.0,3,4.0,8.0,12.0,
        0,1.0,2.0,3.0,1,2.0,4.0,6.0,2,3.0,6.0,9.0,3,4.0,8.0,12.0,3
        """.strip()
        expected_result = np.array([
            [ [0, 1.0,2.0,3.0], [1, 2.0,4.0,6.0], [2, 3.0,6.0,9.0], [3, 4.0,8.0,12.0] ],
            [ [0, 1.0,2.0,3.0], [1, 2.0,4.0,6.0], [2, 3.0,6.0,9.0], [3, 4.0,8.0,12.0] ]
        ])

        for decoder in TestCDecode.decoders:
            mock_stdout = io.BytesIO(bytes(example_input, "ascii"))
            mock_stdout = io.BufferedReader(mock_stdout)
            with self.subTest("Processing mock data output with C decoders", decoder=decoder):
                reader = decoder(np.zeros((2, 4, 4)))
                reader.read(mock_stdout)
                result, time_stopped = reader.get_output()

                self.assertEqual(time_stopped, 3, f"Stop time returned by decoder is incorrect (expected {3}, got {time_stopped})")
                # Test both the passed-in trajectory list and the returned trajectories.
                self.assertTrue(np.all(result == expected_result), "Trajectory data returned by decoder is incorrect")



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import unittest
from example_models import create_decay
from gillespy2.core import Model, StochMLDocument
from gillespy2.core.gillespyError import *
from gillespy2 import ODESolver
from gillespy2 import NumPySSASolver


class TestStochML(unittest.TestCase):

    def test_StochML_from_and_to_model(self):
        model = create_decay()
        stochml = StochMLDocument.from_model(model)
        stochml_model = stochml.to_model('model')
        stochml_model.timespan(model.tspan)
        ode_solver = ODESolver(model=stochml_model)
        stochml_model.run(solver=ode_solver)
        ssa_solver = NumPySSASolver(model=stochml_model)
        stochml_model.run(solver=ssa_solver)



if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import random, math, sys
from collections import OrderedDict
from scipy.integrate import ode, LSODA
import heapq
import numpy as np
import threading
import gillespy2
from gillespy2.solvers.utilities import Tau
from gillespy2.core import GillesPySolver, log, Event, RateRule, AssignmentRule, FunctionDefinition
from gillespy2.core.gillespyError import *
from gillespy2.core.results import Results

eval_globals = math.__dict__

def __piecewise(*args):
    # Eval entry for piecewise functions
    args = list(args)
    solution = None
    if len(args) % 2:
        args.append(True)
    for i, arg in enumerate(args):
        if not i % 2:
            continue
        if arg:
            solution = args[i - 1]
            break
    return solution


def __xor(*args):
    # Eval entry for MathML xor function
    from operator import ixor
    from functools import reduce
    args = list(args)
    return reduce(ixor, args)


eval_globals['false'] = False
eval_globals['true'] = True
eval_globals['piecewise'] = __piecewise
eval_globals['xor'] = __xor


class TauHybridSolver(GillesPySolver):
    """
    This solver uses a root-finding interpretation of the direct SSA method,
    along with ODE solvers to simulate ODE and Stochastic systems
    interchangeably or simultaneously.
    Uses integrators from scipy.integrate.ode to perform calculations used to produce solutions.

    :param model: The model on which the solver will operate.
    :type model: gillespy2.Model
    """
    name = "TauHybridSolver"
    rc = 0
    result = None
    stop_event = None

    def __init__(self, model=None, profile_reactions=False, constant_tau_stepsize=None):
        if model is None:
            raise SimulationError("A model is required to run the simulation.")

        name = 'TauHybridSolver'
        rc = 0
        self.model = copy.deepcopy(model)
        self.is_instantiated = True
        self.profile_reactions = profile_reactions
        self.profile_data = {}
        if self.profile_reactions:
            self.profile_data['time'] = []
            for k in list(self.model.listOfSpecies)+list(self.model.listOfReactions):
                self.profile_data[k] = []
        self.non_negative_species = set()
        for reaction, _ in model.listOfReactions.items():
            for key, value in model.listOfReactions[reaction].reactants.items():
                self.non_negative_species.add(key.name)
            for key, value in model.listOfReactions[reaction].products.items():
                self.non_negative_species.add(key.name)
        self.constant_tau_stepsize = constant_tau_stepsize
        # check if model should skip ODE integration step
        self.pure_discrete = False
        if len(self.model.listOfRateRules)==0:
            self.pure_discrete = True
            for species in self.model.listOfSpecies:
                if self.model.listOfSpecies[species].mode != 'discrete':
                    self.pure_discrete = False
                    break



    def __save_state_to_output(self, curr_time, save_index, curr_state, species, 
                                trajectory, save_times):
        """
        Helper function to save the curr_state to the trajectory output
        """
        
        # Now update the step and trajectories for this step of the simulation.
        # Here we make our final assignments for this step, and begin
        # populating our results trajectory.
           
        num_saves = 0
        while len(save_times)>0 and save_times[0] <= curr_time:
            time = save_times.pop(0) # remove the value from the front of the list
            # if a solution is given for it
            trajectory_index = save_index
            assignment_state = copy.deepcopy(curr_state)
            for s,sname in enumerate(species):
                # Get ODE Solutions
                trajectory[trajectory_index][s + 1] = curr_state[sname]
                # Update Assignment Rules for all processed time points
                if len(self.model.listOfAssignmentRules):
                    # Copy ODE state for assignments
                    assignment_state[sname] = curr_state[sname]
            assignment_state['t'] = time
            for ar in self.model.listOfAssignmentRules.values():
                assignment_value = eval(ar.formula, {**eval_globals, **assignment_state})
                assignment_state[ar.variable] = assignment_value
                trajectory[trajectory_index][species.index(ar.variable.name) + 1] = assignment_value
            num_saves += 1
            save_index += 1
        return save_index


    def __toggle_reactions(self, all_compiled, deterministic_reactions, dependencies,
                            curr_state, det_spec, rr_sets):
        """
        Helper method which is used to convert reaction channels into
        rate rules, and rate rules into reaction channels, as they are switched
        dynamically throughout the simulation based on user-supplied tolerance.
        """

        # initialize variables
        inactive_reactions = all_compiled['inactive_rxns']
        rate_rules = all_compiled['rules']
        rxns = all_compiled['rxns']

        # If the set has changed, reactivate non-determinsitic reactions
        reactivate = []
        for r in inactive_reactions:
            if not r in deterministic_reactions:
                reactivate.append(r)
        for r in reactivate:
            rxns[r] = inactive_reactions.pop(r, None)

        # Deactivate Determinsitic Reactions
        for r in deterministic_reactions:
            if not r in inactive_reactions:
                inactive_reactions[r] = rxns.pop(r, None)

        # floor non-det species
        for s,d in det_spec.items():
            if not d and isinstance(curr_state[s], float):
                curr_state[s] = round(curr_state[s])

        # Check if this reaction set is already compiled and in use:
        if deterministic_reactions in rr_sets.keys():
            return rr_sets[deterministic_reactions]
        else:
        # Otherwise, this is a new determinstic reaction set that must be compiled
            tmp =  self.__create_diff_eqs(deterministic_reactions,
                                            dependencies, rr_sets, rate_rules)
            return tmp

    def __create_diff_eqs(self, comb, dependencies, rr_sets, rate_rules):
        """
        Helper method used to convert stochastic reaction descriptions into
        differential equations, used dynamically throught the simulation.
        """
        diff_eqs = OrderedDict()
        output_rules = copy.deepcopy(rate_rules)

        # Initialize sample dict
        rr_vars = {}
        for n, rr in self.model.listOfRateRules.items():
            rr_vars[rr.variable.name] = n
        for spec in self.model.listOfSpecies:
            if spec in rr_vars.keys():
                diff_eqs[self.model.listOfSpecies[spec]] = self.model.listOfRateRules[rr_vars[spec]].formula
            else:
                diff_eqs[self.model.listOfSpecies[spec]] = '0'

        # loop through each det reaction and concatenate it's diff eq for each species
        for reaction in comb:
            factor = {dep: 0 for dep in dependencies[reaction]}

            for key, value in self.model.listOfReactions[reaction].reactants.items():
                if not key.constant and not key.boundary_condition:
                    factor[key.name] -= value
            for key, value in self.model.listOfReactions[reaction].products.items():
                if not key.constant and not key.boundary_condition:
                    factor[key.name] += value

            for dep in dependencies[reaction]:
                if factor[dep] != 0:
                    if self.model.listOfSpecies[dep].mode == 'continuous':
                        diff_eqs[self.model.listOfSpecies[dep]] += ' + {0}*({1})'.format(factor[dep],
                                                           self.model.listOfReactions[reaction].ode_propensity_function)
                    else:
                        diff_eqs[self.model.listOfSpecies[dep]] += ' + {0}*({1})'.format(factor[dep],
                                                           self.model.listOfReactions[reaction].propensity_function)

        for spec in self.model.listOfSpecies:
            if diff_eqs[self.model.listOfSpecies[spec]] == '0':
                del diff_eqs[self.model.listOfSpecies[spec]]
        # create a dictionary of compiled gillespy2 rate rules
        for spec, rate in diff_eqs.items():
            output_rules[spec] = compile(gillespy2.RateRule(spec, rate).formula, '<string>', 'eval')
        rr_sets[comb] = rate_rules # save values
        return output_rules

    def __flag_det_reactions(self, det_spec, det_rxn, dependencies):
        """
        Helper method used to flag reactions that can be processed
        deterministically without exceeding the user-supplied tolerance.
        """
        # Determine if each rxn would be deterministic apart from other reactions
        prev_state = det_rxn.copy()
        for rxn in self.model.listOfReactions:
            # assume it is deterministic
            det_rxn[rxn] = True
            # iterate through the dependent species of this reaction
            for species in dependencies[rxn]:
                # if any of the dependencies are discrete or (dynamic AND the
                # species itself has not been flagged as deterministic)
                # then allow it to be modelled discretely
                if self.model.listOfSpecies[species].mode == 'discrete':
                    det_rxn[rxn] = False
                    break
                if self.model.listOfSpecies[species].mode == 'dynamic' and det_spec[species] == False:
                    det_rxn[rxn] = False
                    break

        # Create a hashable frozenset of all determinstic reactions
        deterministic_reactions = set()
        for rxn in det_rxn:
            if det_rxn[rxn]:
                deterministic_reactions.add(rxn)
        deterministic_reactions = frozenset(deterministic_reactions)
        return deterministic_reactions

    def __calculate_statistics(self, curr_time, propensities, curr_state, tau_step, det_spec, cv_history={}):
        """
        Calculates Mean, Standard Deviation, and Coefficient of Variance for each
        dynamic species, then set if species can be represented determistically

        NOTE: the argument cv_history should not be passed in, this is modified by the function
        to keep a persistent data set.
        """
        #move configuration to solver init (issue #905)
        history_length = 12 #cite: "Statistical rules of thumb" G Van Belle

        if curr_time==0.0: #re-set cv_history
            for k in list(cv_history.keys()):
                del cv_history[k]

        CV = OrderedDict()
        # calculate CV by estimating the next step
        mn = {species: curr_state[species] for species, value in
              self.model.listOfSpecies.items() if value.mode == 'dynamic'}
        sd = {species: 0 for species, value in
              self.model.listOfSpecies.items() if value.mode == 'dynamic'}

        for r, rxn in self.model.listOfReactions.items():
            for reactant in rxn.reactants:
                if reactant.mode == 'dynamic':
                    mn[reactant.name] -= (propensities[r] * rxn.reactants[reactant])
                    sd[reactant.name] += (propensities[r] * (rxn.reactants[reactant] ** 2))
            for product in rxn.products:
                if product.mode == 'dynamic':
                    mn[product.name] += (propensities[r] * rxn.products[product])
                    sd[product.name] += (propensities[r] * (rxn.products[product] ** 2))
        # Calcuate the derivative based CV
        for species,value in self.model.listOfSpecies.items():
            if value.mode == 'dynamic':
                if mn[species] > 0 and sd[species] > 0:
                    CV[species] = math.sqrt(sd[species]) / mn[species]
                else:
                    CV[species] = 1  # value chosen to guarantee species will be discrete

        # Keep a history of the past CV values, calculate a time-averaged value
        CV_a = OrderedDict() # time-averaged CV (forward derivative based)
        for species,value in self.model.listOfSpecies.items():
            if value.mode == 'dynamic':
                if species not in cv_history:
                    cv_history[species] = []
                cv_history[species].append(CV[species])
                if len(cv_history[species]) > history_length:
                    cv_history[species].pop(0) #remove the first item
                CV_a[species] = sum(cv_history[species])/len(cv_history[species])

        # Select DISCRETE or CONTINOUS mode for each species
        for species in mn:
            prev_det = det_spec[species]
            sref = self.model.listOfSpecies[species]
            if sref.switch_min == 0:
                # Set species to deterministic if CV is less than threshhold
                det_spec[species] = CV_a[species] < sref.switch_tol
            else:
                det_spec[species] = mn[species] > sref.switch_min

        return mn, sd, CV

    @staticmethod
    def __f(t, y, curr_state, species, reactions, rate_rules, propensities,
            y_map, compiled_reactions, active_rr, events, assignment_rules):
        """
        Evaluate the propensities for the reactions and the RHS of the Reactions and RateRules.
        Also evaluates boolean value of event triggers.
        """
        state_change = [0] * len(y_map)
        curr_state['t'] = t
        curr_state['time'] = t
        for item, index in y_map.items():
            if item in assignment_rules:
                curr_state[assignment_rules[item].variable] = eval(assignment_rules[item].formula,
                                                                   {**eval_globals, **curr_state})
            else:
                curr_state[item] = y[index]
        for s, rr in active_rr.items():
            try:
                state_change[y_map[s.name]] += eval(rr, {**eval_globals, **curr_state})
            except ValueError as e:
                pass
        for i, r in enumerate(compiled_reactions):
            propensities[r] = eval(compiled_reactions[r], {**eval_globals, **curr_state})
            state_change[y_map[r]] += propensities[r]
        for event in events:
            triggered = eval(event.trigger.expression, {**eval_globals, **curr_state})
            if triggered:
                state_change[y_map[event]] = 1
        return state_change

    def __find_event_time(self, sol, start, end, index, depth):
        """
        Helper method providing binary search implementation for locating
        precise event times.
        """
        dense_range = np.linspace(start, end, 3)
        mid = dense_range[1]
        if start >= mid or mid >= end or depth == 20: return end
        solutions = np.diff(sol.sol(dense_range)[-len(self.model.listOfEvents) + index])
        bool_res = [x > 0 for x in solutions]

        if bool_res[0]:  # event before mid
            depth += 1
            return self.__find_event_time(sol, dense_range[0],
                                          dense_range[1], index, depth)
        else:  # event after mid
            depth += 1
            return self.__find_event_time(sol, dense_range[1],
                                          dense_range[2], index, depth)

    def __detect_events(self, event_sensitivity, sol, delayed_events,
                        trigger_states, curr_time, curr_state):
        """
        Helper method to locate precise time of event firing.  This method
        first searches for any instance of an event using event_sensitivity to
        determine the granularity of the search.  If an event is detected, a
        binary search is then used to locate the precise time of that event.
        """
        event_times = {}
        dense_range = np.linspace(sol.t[0], sol.t[-1], len(sol.t) * event_sensitivity)
        solutions = np.diff(sol.sol(dense_range))
        for i, e in enumerate(self.model.listOfEvents.values()):
            bool_res = [x > 0 for x in solutions[i - len(self.model.listOfEvents)]]
            curr_state[e.name] = bool_res[-1]
            # Search for changes from False to True in event, record first time
            for y in range(1, len(dense_range) - 1):
                # Check Persistent Delays.  If an event is not designated as
                # persistent, and the trigger expression fails to evaluate as
                # true before assignment is carried out, remove the event from
                # the queue.
                if e.name in trigger_states and not e.trigger.persistent:
                    if not bool_res[y]:
                        delayed_events[i] = delayed_events[-1]  # move to end
                        heapq.heappop(delayed_events)
                        del trigger_states[e.name]
                        curr_state[e.name] = False
                # IF triggered from false to true, refine search
                elif bool_res[y] and dense_range[y] != curr_time and bool_res[y - 1] == 0:
                    event_time = self.__find_event_time(sol, dense_range[y - 1],
                                                        dense_range[y + 1], i, 0)
                    if event_time in event_times:
                        event_times[event_time].append(e)
                    else:
                        event_times[event_time] = [e]
                    break
        return event_times

    def __get_next_step(self, event_times, reaction_times, delayed_events,
                        sim_end, next_tau):
        """
        Helper method to determine the next action to take during simulation,
        and returns that action along with the time that it occurs.
        """

        next_event_trigger = sim_end + 2
        next_delayed_event = sim_end + 3
        curr_time = sim_end

        # event triggers
        if len(event_times):
            next_event_trigger = min(event_times)

        # delayed events
        if len(delayed_events):
            next_delayed_event = delayed_events[0][0]

        # Execute rxns/events
        next_step = {sim_end: 'end', next_tau: 'tau', next_event_trigger: 'trigger', next_delayed_event: 'delay'}

        # Set time to next action
        curr_time = min(sim_end, next_tau, next_event_trigger,
                        next_delayed_event)
        return next_step[curr_time], curr_time

    def __process_queued_events(self, event_queue, trigger_states,
                                curr_state, det_spec):
        """
        Helper method which processes the events queue. Method is primarily for
        evaluating assignments at the designated state (trigger time or event
        time).
        """
        # Process all queued events
        events_processed = []
        pre_assignment_state = curr_state.copy()
        while event_queue:
            # Get events in priority order
            fired_event = self.model.listOfEvents[heapq.heappop(event_queue)[1]]
            events_processed.append(fired_event)
            if fired_event.name in trigger_states:
                assignment_state = trigger_states[fired_event.name]
                del trigger_states[fired_event.name]
            else:
                assignment_state = pre_assignment_state
            for a in fired_event.assignments:
                # Get assignment value
                assign_value = eval(a.expression, eval_globals, assignment_state)
                # Update state of assignment variable
                curr_state[a.variable.name] = assign_value

        return events_processed

    def __handle_event(self, event, curr_state, curr_time, event_queue,
                       trigger_states, delayed_events):
        """
        Helper method providing logic for updating states based on assignments.
        """
        # Fire trigger time events immediately
        if event.delay is None:
            heapq.heappush(event_queue, (eval(event.priority), event.name))
        # Queue delayed events
        else:
            curr_state['t'] = curr_time
            curr_state['time'] = curr_time
            execution_time = curr_time + eval(event.delay, {**eval_globals, **curr_state})
            curr_state[event.name] = True
            heapq.heappush(delayed_events, (execution_time, event.name))
            if event.use_values_from_trigger_time:
                trigger_states[event.name] = curr_state.copy()
            else:
                trigger_states[event.name] = curr_state

    def __check_t0_events(self, initial_state):
        """
        Helper method for firing events who reach a trigger condition at start
        of simulation, time == 0.
        """
        # Check Event State at t==0
        species_modified_by_events = []
        t0_delayed_events = {}
        for e in self.model.listOfEvents.values():
            if not e.trigger.value:
                t0_firing = eval(e.trigger.expression, {**eval_globals, **initial_state})
                if t0_firing:
                    if e.delay is None:
                        for a in e.assignments:
                            initial_state[a.variable.name] = eval(a.expression, {**eval_globals, **initial_state})
                            species_modified_by_events.append(a.variable.name)
                    else:
                        execution_time = eval(e.delay, {**eval_globals, **initial_state})
                        t0_delayed_events[e.name] = execution_time
        return t0_delayed_events, species_modified_by_events

    def __update_stochastic_rxn_states(self, propensities, tau_step, compiled_reactions, curr_state, only_update=None):
        """
        Helper method for updating the state of stochastic reactions.

        if 'only_update' is set to a reaction name, it will only reset that reaction, and only one firing
        """

        rxn_count = OrderedDict()
        species_modified = OrderedDict()
        # Update stochastic reactions

        for rxn in compiled_reactions:
            rxn_count[rxn] = 0
            if only_update is not None:  # for a single SSA step
                if rxn == only_update:
                    curr_state[rxn] = math.log(np.random.uniform(0, 1)) #set this value, needs to be <0
                    rxn_count[rxn] = 1
            elif curr_state[rxn] > 0:
                rxn_count[rxn] = 1 + np.random.poisson(curr_state[rxn])
                curr_state[rxn] = math.log(np.random.uniform(0, 1))

            if rxn_count[rxn]:
                for reactant in self.model.listOfReactions[rxn].reactants:
                    species_modified[reactant.name] = True
                    curr_state[reactant.name] -= self.model.listOfReactions[rxn].reactants[reactant] * rxn_count[rxn]
                for product in self.model.listOfReactions[rxn].products:
                    species_modified[product.name] = True
                    curr_state[product.name] += self.model.listOfReactions[rxn].products[product] * rxn_count[rxn]
        return species_modified, rxn_count

    def __integrate_constant(self, integrator_options, curr_state, y0, curr_time,
                    propensities, y_map, compiled_reactions,
                    active_rr, event_queue,
                    delayed_events, trigger_states,
                    event_sensitivity, tau_step ):
        """
        Integreate one step, skip ODE integrator as all species are 
        discrete.
        """
        tau_step = max(1e-6, tau_step)
        next_tau = curr_time + tau_step
        prev_time = curr_time
        curr_state['t'] = curr_time
        curr_state['time'] = curr_time
        
        event_times = {}
        reaction_times = []
        next_step, curr_time = self.__get_next_step(event_times, reaction_times,
                                                    delayed_events,
                                                    self.model.tspan[-1], next_tau)
        curr_state['t'] = curr_time

        tau_step2 = curr_time - prev_time

        for rxn in compiled_reactions:
            curr_state[rxn] = curr_state[rxn] + propensities[rxn]*tau_step2

        return curr_time, tau_step2



    def __integrate(self, integrator_options, curr_state, y0, curr_time,
                    propensities, y_map, compiled_reactions,
                    active_rr, event_queue,
                    delayed_events, trigger_states,
                    event_sensitivity, tau_step, det_rxn ):
        """
        Helper function to perform the ODE integration of one step.  This
        method uses scipy.integrate.LSODA to get simulation data, and
        determines the next stopping point of the simulation. The state is
        updated and returned to __simulate along with curr_time and the
        solution object.
        """
        use_const = False
        if self.pure_discrete:
            use_const = True
        elif len(self.model.listOfRateRules)==0:
            use_const = True
            for rxn in det_rxn:
                if det_rxn[rxn]:
                    use_const = False
                    break

        if use_const:
            return self.__integrate_constant(integrator_options, curr_state, y0, curr_time,
                            propensities, y_map, compiled_reactions, active_rr, event_queue,
                            delayed_events, trigger_states, event_sensitivity, tau_step)

        from functools import partial
        events = self.model.listOfEvents.values()
        dense_output = False
        int_args = [curr_state, self.model.listOfSpecies, self.model.listOfReactions,
                    self.model.listOfRateRules,
                    propensities, y_map,
                    compiled_reactions,
                    active_rr,
                    events,
                    self.model.listOfAssignmentRules]
        rhs = lambda t, y: TauHybridSolver.__f(t, y, *int_args)
        if 'min_step' in integrator_options:
            tau_step = max(integrator_options['min_step'], tau_step)
        else:
            tau_step = max(1e-6, tau_step)
        #next_tau = curr_time + tau_step
        last_curr_time = curr_time

        # Set curr time to next time a change occurs in the system outside of
        # the standard ODE process.  Determine what kind of change this is,
        # and set the curr_time of simulation to restart simulation after
        # making the appropriate state changes.
        event_times = {}
        reaction_times = []
        next_step, curr_time = self.__get_next_step(event_times, reaction_times,
                                                    delayed_events,
                                                    self.model.tspan[-1], curr_time + tau_step)
        curr_state['t'] = curr_time
        curr_state['time'] = curr_time
        actual_tau_step = last_curr_time - curr_time

        # Integrate until end or tau is reached
        loop_count = 0
        sol = LSODA(rhs, last_curr_time, y0, curr_time)
        counter = 0
        while sol.t < curr_time:
            counter += 1
            sol.step()


        # Update states of all species based on changes made to species through
        # ODE processes.  This will update all species whose mode is set to
        # 'continuous', as well as 'dynamic' mode species which have been
        # flagged as deterministic.
       
        for spec_name, species in self.model.listOfSpecies.items():
            if not species.constant:
                curr_state[spec_name] = sol.y[y_map[spec_name]]

        # Search for precise event times
        '''
        if len(model.listOfEvents):
            event_times = self.__detect_events(event_sensitivity, sol, delayed_events,
                                               trigger_states, curr_time, curr_state)
        else:
            event_times = {}

        # Get next tau time
        reaction_times = []
        '''

        # Stochastic Reactions are also fired through a root-finding method
        # which mirrors the standard SSA probability.  Since we are using
        # Tau-Leaping, rather than firing reactions based on a direct poisson
        # distribution from the propensity, we use that same poisson
        # distribution to select a random negative offset for the reaction
        # state, then integrate forward along with any deterministic
        # reactions/species for a designated time (tau).  Root crossings
        # satisfy the SSA firing process, and because multiple reactions can be
        # fired in a single Tau step, a new random number will be generated and
        # added (representing a single-firing each time) until the state value
        # of the reaction is once again negative.
        for rxn in compiled_reactions:
            curr_state[rxn] = sol.y[y_map[rxn]]

        # In the case that a major change occurs to the system (outside of the
        # standard ODE process) is caused by an event trigger, we then examine
        # the event which was triggered.  if Event assignments are carried out
        # immediately, we add them to a heap in Event Priority order to be
        # immediately dealt with.  If an Event contains a delay, we can now set
        # the time of event assignment execution by evaluating the event delay
        # in the current state.
        if next_step == 'trigger':
            for event in event_times[curr_time]:
                self.__handle_event(event, curr_state, curr_time,
                                    event_queue, trigger_states, delayed_events)
        # In the case that a major change occurs to the system (outside of the
        # standard ODE process) is caused by a delayed event which has now
        # reached the designated time of execution, we add all currently
        # designated event assignments to a heap to be processed in Event
        # Priority order.
        elif next_step == 'delay':
            event = heapq.heappop(delayed_events)
            heapq.heappush(event_queue, (eval(self.model.listOfEvents[event[1]].priority), event[1]))

        return curr_time, actual_tau_step


    def __simulate_negative_state_check(self, curr_state):
            # check each species to see if they are negative
            for s in self.non_negative_species:
                if curr_state[s] < 0:
                    raise SimulationError(f"Negative State detected at begining of step." \
                    " Species involved in reactions can not be negative.")

    def __simulate_invalid_state_check(self, species_modified, curr_state, compiled_reactions):
        invalid_state = False
        err_message=""
        # check each species to see if they are negative
        for s in species_modified.keys():
            if curr_state[s] < 0:
                invalid_state = True
                err_message += f"'{s}' has negative state '{curr_state[s]}'"
        return (invalid_state, err_message)

    def __simulate(self, integrator_options, curr_state, y0, curr_time,
                   propensities, species, parameters, compiled_reactions,
                   active_rr, y_map, trajectory, save_times, save_index,
                   delayed_events, trigger_states, event_sensitivity,
                   tau_step, debug, det_spec, det_rxn):
        """
        Function to process simulation until next step, which can be a
        stochastic reaction firing, an event trigger or assignment, or end of
        simulation.

        :param curr_state: Contains all state variables for system at current time
        :type curr_state: dict

        :param curr_time: Represents current time
        :type curr_time: float

        :param save_times: Currently unreached save points
        :type save_times: list

        :returns: curr_state, curr_time, save_times, sol
        """

        # first check if we have a valid state:
        self.__simulate_negative_state_check(curr_state)
        if curr_time == 0.0:
            # save state at beginning of simulation
            save_index = self.__save_state_to_output(
                curr_time, save_index, curr_state, species, trajectory, save_times
            )

        event_queue = []
        prev_y0 = copy.deepcopy(y0)
        prev_curr_state = copy.deepcopy(curr_state)
        prev_curr_time = curr_time
        loop_count = 0
        invalid_state = False


        starting_curr_state = copy.deepcopy(curr_state)
        starting_propensities = copy.deepcopy(propensities)
        starting_tau_step=tau_step
        species_modified=None
        rxn_count=None
        loop_err_message=""
        if self.profile_reactions:
            self.profile_data['time'].append(curr_time)
            for k in list(self.model.listOfSpecies)+list(self.model.listOfReactions):
                self.profile_data[k].append(curr_state[k])


        # check each reaction to see if it is >=0. If we have taken a single SSA step, this could be >0 for the non-selected reactions, check if propensity is zero and reset if so
        for r in compiled_reactions.keys():
            if curr_state[r] >= 0 and propensities[r] == 0:
                curr_state[r] = math.log(np.random.uniform(0, 1))

        curr_time, actual_tau_step = self.__integrate(integrator_options, curr_state,
                                          y0, curr_time, propensities, y_map,
                                          compiled_reactions,
                                          active_rr,
                                          event_queue,
                                          delayed_events,
                                          trigger_states,
                                          event_sensitivity,
                                          tau_step, det_rxn
                                          )

        species_modified,rxn_count = self.__update_stochastic_rxn_states(propensities, actual_tau_step, compiled_reactions, curr_state)

        # Occasionally, a tau step can result in an overly-aggressive
        # forward step and cause a species population to fall below 0,
        # which would result in an erroneous simulation.
        # We estimate the time to the first
        # stochatic reaction firing (assume constant propensities) and
        # simulate the ODE system until that time, fire that reaction
        # and continue the simulation.
        (invalid_state, invalid_err_message) = self.__simulate_invalid_state_check(species_modified, curr_state, compiled_reactions)

        if invalid_state:
            invalid_state = False
            # Redo this step, with a smaller Tau such that a single SSA reaction occurs
            y0 = copy.deepcopy(prev_y0)
            curr_state_after = copy.deepcopy(curr_state)
            curr_state = copy.deepcopy(prev_curr_state)
            floored_curr_state = copy.deepcopy(prev_curr_state)
            propensities_after = copy.deepcopy(propensities)
            propensities = copy.deepcopy(starting_propensities)
            floored_propensities = copy.deepcopy(starting_propensities)
            curr_time = prev_curr_time

            # floored propensites
            for i, s in enumerate(self.model.listOfSpecies):
                floored_curr_state[s] = math.floor(floored_curr_state[s])
            saved_curr_state = copy.deepcopy(curr_state)
            curr_state = floored_curr_state
            for i, r in enumerate(compiled_reactions):
                try:
                    floored_propensities[r] = eval(compiled_reactions[r], {**eval_globals, **curr_state})
                except Exception as e:
                    raise SimulationError('Error calculation propensity for {0}.\nReason: {1}\nfloored_propensities={2}\ncompiled_reactions={3}'.format(r, e, floored_propensities,compiled_reactions))
            curr_state = saved_curr_state

            rxn_times = OrderedDict()
            min_tau = None
            rxn_selected = None
            for rname,rcnt in rxn_count.items():
                if floored_propensities[rname] > 0.0:
                    # estimate the zero crossing time
                    rxn_times[rname] = -1* curr_state[rname] / propensities[rname]
                    if min_tau is None or min_tau > rxn_times[rname]:
                        min_tau = rxn_times[rname]
                        rxn_selected = rname
            if rxn_selected is None: raise SimulationError(f"Negative State detected in step, and no reaction found to fire. error_message={invalid_err_message}")

            tau_step = min_tau #estimated time to the first stochatic reaction

            curr_time, actual_tau_step = self.__integrate(integrator_options, curr_state,
                                          y0, curr_time, propensities, y_map,
                                          compiled_reactions,
                                          active_rr,
                                          event_queue,
                                          delayed_events,
                                          trigger_states,
                                          event_sensitivity,
                                          tau_step, det_rxn
                                          )

            # only update the selected reaction
            first_rxn_count = copy.deepcopy(rxn_count)
            first_err_message = invalid_err_message
            species_modified,rxn_count = self.__update_stochastic_rxn_states(propensities, actual_tau_step, compiled_reactions, curr_state, only_update=rxn_selected)

            (invalid_state, invalid_err_message) = self.__simulate_invalid_state_check(species_modified, curr_state, compiled_reactions)
            if invalid_state:
                raise SimulationError(f"Negative State detected in step, after single SSA step. error_message={invalid_err_message}")


        save_index = self.__save_state_to_output(curr_time, save_index, curr_state, species, trajectory, save_times)

        events_processed = self.__process_queued_events(event_queue, trigger_states, curr_state, det_spec)

        # Finally, perform a final check on events after all non-ODE assignment
        # changes have been carried out on model.
        event_cycle = True
        while event_cycle:
            event_cycle = False
            for i, e in enumerate(self.model.listOfEvents.values()):
                triggered = eval(e.trigger.expression, {**eval_globals, **curr_state})
                if triggered and not curr_state[e.name]:
                    curr_state[e.name] = True
                    self.__handle_event(e, curr_state, curr_time,
                                        event_queue, trigger_states, delayed_events)
                    event_cycle = True
                elif not triggered:
                    curr_state[e.name] = False

        events_processed = self.__process_queued_events(event_queue, trigger_states, curr_state, det_spec)

        return curr_state, curr_time, save_times, save_index

    def __set_seed(self, seed):
        # Set seed if supplied
        if seed is not None:
            if not isinstance(seed, int):
                seed = int(seed)
            if seed > 0:
                np.random.seed(seed)
            else:
                raise ModelError('seed must be a positive integer')

    def __set_recommended_ode_defaults(self, integrator_options):
        """
        Set some ODE solver defaults.  These values are chosen based on the
        precision required to successfully complete the SBML Test suite.
        """

        if 'rtol' not in integrator_options:
            integrator_options['rtol'] = 1e-9
        if 'atol' not in integrator_options:
            integrator_options['atol'] = 1e-12
        if 'max_step' not in integrator_options:
            integrator_options['max_step'] = 0.25

    def __compile_all(self):
        """
        Compile all run-time evaluables to enhance performance.
        """
        compiled_reactions = OrderedDict()
        for i, r in enumerate(self.model.listOfReactions):
            compiled_reactions[r] = compile(self.model.listOfReactions[r].propensity_function, '<string>',
                                            'eval')

        compiled_rate_rules = OrderedDict()
        for i, rr in enumerate(self.model.listOfRateRules.values()):
            if isinstance(rr.variable, str):
                compiled_rate_rules[self.model.listOfSpecies[rr.variable]] = compile(
                                                            rr.formula, '<string>', 'eval')
            else:
                compiled_rate_rules[rr.variable] = compile(rr.formula, '<string>', 'eval')
        compiled_inactive_reactions = OrderedDict()

        compiled_propensities = copy.deepcopy(compiled_reactions)

        return compiled_reactions, compiled_rate_rules, compiled_inactive_reactions, compiled_propensities

    def __initialize_state(self, curr_state, debug):
        """
        Initialize curr_state for each trajectory.
        """

        # intialize parameters to current state
        for p in self.model.listOfParameters:
            curr_state[p] = self.model.listOfParameters[p].value

        # initialize species population state
        for s in self.model.listOfSpecies:
            curr_state[s] = self.model.listOfSpecies[s].initial_value

        # Set reactions to uniform random number
        for i, r in enumerate(self.model.listOfReactions):
            curr_state[r] = math.log(np.random.uniform(0, 1))
            if debug:
                print("Setting Random number ", curr_state[r], " for ", self.model.listOfReactions[r].name)

        # Initialize event last-fired times to 0
        for e_name in self.model.listOfEvents:
            curr_state[e_name] = 0

        sanitized_species = self.model.sanitized_species_names()
        sanitized_parameters = self.model.sanitized_parameter_names()
        for fd in self.model.listOfFunctionDefinitions.values():
            sanitized_function = fd.sanitized_function(sanitized_species, sanitized_parameters)
            curr_state[fd.name] = eval(f"lambda {', '.join(fd.args)}: {sanitized_function}", eval_globals)

        for ar in self.model.listOfAssignmentRules.values():
            if ar.variable in self.model.listOfSpecies:
                continue
            curr_state[ar.variable] = ar.formula

    def __map_state(self, species, parameters, compiled_reactions, events, curr_state):
        """
        Creates the start state vector for integration and provides a
        dictionary map to it's elements.
        """
        y_map = OrderedDict()
        # Build integration start state
        y0 = [0] * (len(species) + len(parameters) + len(compiled_reactions) + len(events))
        for i, spec in enumerate(species):
            if isinstance(curr_state[spec], str):
                y0[i] = eval(curr_state[spec], {**eval_globals, **curr_state})
            else:
                y0[i] = curr_state[spec]
            y_map[spec] = i
        for i, param in enumerate(parameters):
            y0[i + len(species)] = curr_state[param]
            y_map[param] = i + len(species)
        for i, rxn in enumerate(compiled_reactions):
            y0[i + len(species) + len(parameters)] = curr_state[rxn]
            y_map[rxn] = i + len(species) + len(parameters)
        for i, event in enumerate(events.values()):
            y0[i + len(species) + len(parameters) + len(compiled_reactions)] = curr_state[event.name]
            y_map[event] = i + len(species) + len(parameters) + len(compiled_reactions)
        return y0, y_map

    @classmethod
    def get_solver_settings(cls):
        """
        Returns a list of arguments supported by tau_hybrid_solver.run.
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        :rtype: tuple
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'seed', 'debug', 'profile', 'tau_tol',
                'event_sensitivity', 'integrator_options', 'timeout')

    @classmethod
    def get_supported_features(cls):
        return {
            Event,
            RateRule,
            AssignmentRule,
            FunctionDefinition,
        }

    def run(self=None, model=None, t=None, number_of_trajectories=1, increment=None, seed=None,
            debug=False, profile=False, tau_tol=0.03, event_sensitivity=100,integrator_options={},
            live_output=None, live_output_options={}, timeout=None, **kwargs):
        """
        Function calling simulation of the model. This is typically called by the run function in GillesPy2 model
        objects and will inherit those parameters which are passed with the model as the arguments this run function.

        :param model: The model on which the solver will operate. (Deprecated)
        :type model: gillespy2.Model

        :param t: Simulation run time.
        :type t: int or float

        :param number_of_trajectories: Number of trajectories to simulate. By default number_of_trajectories = 1.
        :type number_of_trajectories: int

        :param increment: Save point increment for recording data.
        :type increment: float

        :param seed: The random seed for the simulation. Optional, defaults to None.
        :type seed: int

        :param debug: Set to True to provide additional debug information about the simulation.
        :type debug: bool

        :param profile: Set to True to provide information about step size (tau) taken at each step.
        :type profile: bool

        :param tau_tol: Tolerance level for Tau leaping algorithm.  Larger tolerance values will
        result in larger tau steps. Default value is 0.03.
        :type tau_tol: float

        :param event_sensitivity: Number of data points to be inspected between integration
            steps/save points for event detection. Default event_sensitivity = 100
        :type event_sensitivity: int

        :param integrator_options:  contains options to the scipy integrator. by default, this includes
            rtol=1e-9 and atol=1e-12.  for a list of options,
            see https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.LSODA.html.
            Example use: {max_step : 0, rtol : .01}
        :type integrator_options: dict

        :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :type live_output: str

        :param live_output_options: contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each display
        :type live_output_options:  dict

        :param timeout: If set, if simulation takes longer than timeout, will exit.
        :type timeout: int
       
        :returns: A result object containing the results of the simulation.
        :rtype: gillespy2.Results
        """
        from gillespy2 import log

        if self is None:
            # Post deprecation block
            # raise SimulationError("TauHybridSolver must be instantiated to run the simulation")
            # Pre deprecation block
            log.warning(
                """
                `gillespy2.Model.run(solver=TauHybridSolver)` is deprecated.

                You should use `gillespy2.Model.run(solver=TauHybridSolver(model=gillespy2.Model))
                Future releases of GillesPy2 may not support this feature.
                """
            )
            self = TauHybridSolver(model=model)

        if model is not None:
            log.warning('model = gillespy2.model is deprecated. Future releases '
                        'of GillesPy2 may not support this feature.')
        if self.model is None:
            if model is None:
                raise SimulationError("A model is required to run the simulation.")
            self.model = copy.deepcopy(model)

        self.model.compile_prep()
        self.validate_model(self.model, model)
        self.validate_sbml_features(model=self.model)


        self.validate_tspan(increment=increment, t=t)
        if increment is None:
            increment = self.model.tspan[-1] - self.model.tspan[-2]
        if t is None:
            t = self.model.tspan[-1]

        if timeout is not None and timeout > 0:
            for i, s in enumerate(list(self.model._listOfSpecies.keys())):
                # Solve_ivp doesn't return any results until it's finished solving so timing out early only slows
                # the solver.
                if self.model.listOfSpecies[s].mode == 'continuous':
                    timeout = 0
                    log.warning('timeouts not supported by continuous species.')
                    break
                elif self.model.listOfSpecies[s].mode == 'dynamic':
                    log.warning('timeouts not fully supported by dynamic species. If timeout is triggered during'
                                ' integration, total solve time could be longer than expected.')
                    break

        self.stop_event = threading.Event()

        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

        if timeout is not None and timeout <= 0:
            timeout = None

        if debug:
            print("t = ", t)
            print("increment = ", increment)

        if len(self.model.listOfEvents):
            self.__set_recommended_ode_defaults(integrator_options)
        self.__set_seed(seed)

        species = list(self.model._listOfSpecies.keys())
        number_species = len(species)

        initial_state = OrderedDict()
        self.__initialize_state(initial_state, debug)
        initial_state['vol'] = self.model.volume
        initial_state['t'] = 0

        # create numpy array for timeline
        timeline = np.linspace(0, t, int(round(t / increment + 1)))
        self.model.tspan = timeline

        # create numpy matrix to mark all state data of time and species
        trajectory_base = np.zeros((number_of_trajectories, timeline.size, number_species + 1))

        # copy time values to all trajectory row starts
        trajectory_base[:, :, 0] = timeline

        # copy initial populations to base
        spec_modes = ['continuous', 'dynamic', 'discrete', None]
        for i, s in enumerate(species):
            if self.model.listOfSpecies[s].mode is None:
                self.model.listOfSpecies[s].mode = 'dynamic'

            if self.model.listOfSpecies[s].mode not in spec_modes:
                raise SpeciesError('Species mode can only be \'continuous\', \'dynamic\',\'discrete\', or '
                                   '\'unspecified(default to dynamic)\'.')
            trajectory_base[:, 0, i + 1] = initial_state[s]

        # curr_time and curr_state are list of len 1 so that __run receives reference
        curr_time = [0]  # Current Simulation Time
        curr_state = [None]
        live_grapher = [None]

        sim_thread = threading.Thread(target=self.___run,
                                      args=(curr_state, curr_time, timeline, trajectory_base, initial_state,
                                            live_grapher,), kwargs={'t': t,
                                                                    'number_of_trajectories': number_of_trajectories,
                                                                    'increment': increment, 'seed': seed,
                                                                    'debug': debug, 'profile': profile,
                                                                    'tau_tol': tau_tol,
                                                                    'event_sensitivity': event_sensitivity,
                                                                    'integrator_options': integrator_options})
        try:
            sim_thread.start()

            if live_output is not None:
                live_output_options['type'] = live_output

                import gillespy2.core.liveGraphing
                gillespy2.core.liveGraphing.valid_graph_params(live_output_options)

                if live_output_options['type'] == "graph":
                    for i, s in enumerate(list(self.model._listOfSpecies.keys())):

                        if self.model.listOfSpecies[s].mode == 'continuous':
                            log.warning('display \"type\" = \"graph\" not recommended with continuous species. '
                                        'Try display \"type\" = \"text\" or \"progress\".')
                            break

                live_grapher[0] = gillespy2.core.liveGraphing.LiveDisplayer(self.model, timeline, number_of_trajectories,
                                                                            live_output_options)
                display_timer = gillespy2.core.liveGraphing.RepeatTimer(live_output_options['interval'],
                                                                        live_grapher[0].display,
                                                                        args=(curr_state, curr_time, trajectory_base, live_output))
                display_timer.start()

            sim_thread.join(timeout=timeout)

            if live_grapher[0] is not None:
                display_timer.cancel()

            self.stop_event.set()
            while self.result is None: pass
        except:
            pass
        if hasattr(self, 'has_raised_exception'):
            raise self.has_raised_exception 

        return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, curr_state, curr_time, timeline, trajectory_base, initial_state, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, seed=None,
               debug=False, profile=False, tau_tol=0.03, event_sensitivity=100,
               integrator_options={}, **kwargs):
        try:
            self.__run(curr_state, curr_time, timeline, trajectory_base, initial_state, live_grapher, t,
                       number_of_trajectories, increment, seed, debug,
                       profile, tau_tol, event_sensitivity,
                       integrator_options, **kwargs)
        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, curr_state, curr_time, timeline, trajectory_base, initial_state, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, seed=None,
              debug=False, profile=False,
              tau_tol=0.03, event_sensitivity=100,
              integrator_options={}, **kwargs):

        # create mapping of species dictionary to array indices
        species_mappings = self.model._listOfSpecies
        species = list(species_mappings.keys())
        parameter_mappings = self.model._listOfParameters
        parameters = list(parameter_mappings.keys())
        number_species = len(species)

        t0_delayed_events, species_modified_by_events = self.__check_t0_events(initial_state)

        # Create deterministic tracking data structures
        det_spec = {species: value.mode != 'discrete' for (species, value) in self.model.listOfSpecies.items()}

        det_rxn = {rxn: False for (rxn, value) in self.model.listOfReactions.items()}

        # Determine if entire simulation is ODE or Stochastic, in order to
        # avoid unnecessary calculations during simulation
        simulation_data = []

        dependencies = OrderedDict()

        # If considering deterministic changes, create dependency data
        # structure for creating diff eqs later
        for reaction in self.model.listOfReactions:
            dependencies[reaction] = set()
            [dependencies[reaction].add(reactant.name) for reactant in self.model.listOfReactions[reaction].reactants]
            [dependencies[reaction].add(product.name) for product in self.model.listOfReactions[reaction].products]

        # Main trajectory loop
        for trajectory_num in range(number_of_trajectories):

            if self.stop_event.is_set():
                print('exiting')
                self.rc = 33
                break

            # For multi trajectories, live_grapher needs to be informed of trajectory increment
            if live_grapher[0] is not None:
                live_grapher[0].increment_trajectory(trajectory_num)

            trajectory = trajectory_base[trajectory_num]  # NumPy array containing this simulation's results
            propensities = OrderedDict()  # Propensities evaluated at current state

            curr_state[0] = initial_state.copy()
            curr_time[0] = 0  # Current Simulation Time

            for i, r in enumerate(self.model.listOfReactions):
                curr_state[0][r] = math.log(np.random.uniform(0, 1))
                if debug:
                    print("Setting Random number ", curr_state[0][r], " for ", self.model.listOfReactions[r].name)

            end_time = self.model.tspan[-1]  # End of Simulation time
            entry_pos = 1
            data = OrderedDict()  # Dictionary for results
            data['time'] = timeline  # All time entries
            save_index = 0

            # Record Highest Order reactant for each reaction and set error tolerance
            HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, critical_threshold = Tau.initialize(self.model, tau_tol)

            # One-time compilations to reduce time spent with eval
            compiled_reactions, compiled_rate_rules, compiled_inactive_reactions, compiled_propensities = \
                self.__compile_all()
            all_compiled = OrderedDict()
            all_compiled['rxns'] = compiled_reactions
            all_compiled['inactive_rxns'] = compiled_inactive_reactions
            all_compiled['rules'] = compiled_rate_rules

            save_times = list(np.copy(self.model.tspan))
            delayed_events = []
            trigger_states = {}

            # Handle delayed t0 events
            for state in trigger_states.values():
                if state is None: state = curr_state[0]
            for ename, etime in t0_delayed_events.items():
                curr_state[0][ename] = True
                heapq.heappush(delayed_events, (etime, ename))
                if self.model.listOfEvents[ename].use_values_from_trigger_time:
                    trigger_states[ename] = curr_state[0].copy()
                else:
                    trigger_states[ename] = curr_state[0]

            # Each save step
            while curr_time[0] < self.model.tspan[-1]:

                if self.stop_event.is_set():
                    self.rc = 33
                    break
                # Get current propensities
                for i, r in enumerate(self.model.listOfReactions):
                    try:
                        propensities[r] = eval(compiled_propensities[r], eval_globals, curr_state[0])
                        if curr_state[0][r] > 0 and propensities[r]==0:
                            # This is an edge case, that might happen after a single SSA step.
                            curr_state[0][r] = math.log(np.random.uniform(0, 1))
                    except Exception as e:
                        raise SimulationError('Error calculation propensity for {0}.\nReason: {1}\ncurr_state={2}'.format(r, e, curr_state))

                # Calculate Tau statistics and select a good tau step
                if self.constant_tau_stepsize is None:
                    tau_step = Tau.select(HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, tau_tol, critical_threshold, self.model, propensities, curr_state[0], curr_time[0], save_times[0])
                else:
                    tau_step = self.constant_tau_stepsize
                if debug:
                    X = ''
                    for k,v in curr_state[0].items():
                        X+= f'{k}:{v:.2f} '
                    print(f"t={curr_time[0]} tau={tau_step} curr_state={X}")


                # Process switching if used
                mn, sd, CV = self.__calculate_statistics(curr_time[0], propensities, curr_state[0], tau_step, det_spec)

                # Calculate sd and CV for hybrid switching and flag deterministic reactions
                deterministic_reactions = self.__flag_det_reactions(det_spec, det_rxn, dependencies)

                if debug:
                    print('mean: {0}'.format(mn))
                    print('standard deviation: {0}'.format(sd))
                    print('CV: {0}'.format(CV))
                    print('det_spec: {0}'.format(det_spec))
                    print('det_rxn: {0}'.format(det_rxn))

                # Set active reactions and rate rules for this integration step
                rr_sets = {frozenset() : compiled_rate_rules} # base rr set
                active_rr = self.__toggle_reactions(all_compiled, deterministic_reactions,
                                                        dependencies, curr_state[0], det_spec, rr_sets)

                # Create integration initial state vector
                y0, y_map = self.__map_state(species, parameters,
                                             compiled_reactions, self.model.listOfEvents, curr_state[0])

                # Run simulation to next step
                curr_state[0], curr_time[0], save_times, save_index = self.__simulate(integrator_options,
                                                                               curr_state[0], y0, curr_time[0],
                                                                               propensities, species,
                                                                               parameters, compiled_reactions,
                                                                               active_rr, y_map,
                                                                               trajectory, save_times, save_index,
                                                                               delayed_events,
                                                                               trigger_states,
                                                                               event_sensitivity, tau_step,
                                                                               debug, det_spec, det_rxn)

            # End of trajectory, format results
            data = {'time': timeline}
            for i in range(number_species):
                data[species[i]] = trajectory[:, i + 1]
            simulation_data.append(data)

        self.result = simulation_data
        return simulation_data, self.rc



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from gillespy2.core.model import Model


class TestEmptyModel(unittest.TestCase):
    def setUp(self):
        def create_empty_model(parameter_values=None):
            return Model()

        self.model = create_empty_model()

    def test_model_creation(self):
        name = self.model.name
        self.assertEqual(name, "", msg="Unexpected value: {}".format(name))


if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import json
import threading

from math import floor

from gillespy2.core.results import common_rgb_values
from gillespy2.core.gillespyError import SimulationError
from gillespy2.core import log

class CRepeatTimer(threading.Timer):
    """
    Threading timer which repeatedly calls the given function instead of simply ending.

    Used for C solver live graphing support
    """
    pause = False

    def run(self):
        _ = str.join('', [*self.args[1]])
        while not self.finished.wait(self.interval):
            args = self.args[0].get()
            self.function(*args, **self.kwargs)

        if not self.pause:
            args = self.args[0].get()
            self.kwargs['finished'] = True
            self.function(*args, **self.kwargs)


class RepeatTimer(threading.Timer):
    """
    Threading timer which repeatedly calls the given function instead of simply ending
    """
    pause = False

    def run(self):
        _ = str.join('', [*self.args[3]])
        self.args = self.args[:3]
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)

        if not self.pause:
            self.kwargs['finished'] = True
            self.function(*self.args, **self.kwargs)



def display_types():
    '''
    Get the list of supported display types.

    :returns: Supported display types.
    :rtype: list
    '''
    return ["graph", "text", "progress"]


def valid_graph_params(live_output_options):
    '''
    Validated the live output options.

    :param live_output_options: Options to be validated.
    :type live_output_options: dict

    :raises SimulationError: If the display type is invalid.
    '''
    if live_output_options['type'] not in ['progress', 'graph', 'text']:
        raise SimulationError("Invalid input to 'live_output', please check spelling and ensure input is"
                              " lower case.")
    if 'interval' not in live_output_options:
        live_output_options['interval'] = 1
    elif live_output_options['interval'] < 0:
        message = f"In LiveGraphing live_output_options, got 'interval' = '{live_output_options['interval']}'."
        message += " setting interval = 1"
        log.warning(message)
        live_output_options['interval'] = 1

    if live_output_options['type'] == "graph" and live_output_options['interval'] < 1:
        message = f"In LiveGraphing live_output_options, got 'interval' = '{live_output_options['interval']}'."
        message += "Consider using an interval >= 1 when displaying graphs."
        log.warning(message)

    if 'clear_output' not in live_output_options:
        if live_output_options['type'] == "graph" or live_output_options['type'] == "progress":
            live_output_options['clear_output'] = True
        else:
            live_output_options['clear_output'] = False

    if 'file_path' not in live_output_options:
        live_output_options['file_path'] = None
    elif live_output_options['type'] == "graph" and live_output_options['file_path'] is not None:
        live_output_options['type'] = "figure"


class LiveDisplayer():
    """
    holds information required for displaying information when live_output = True
    """

    def __init__(self, model=None, timeline=None, number_of_trajectories=1, live_output_options={}, resume=False):

        self.display_type = live_output_options['type']
        self.display_interval = live_output_options['interval']
        self.file_path = live_output_options['file_path']
        self.model = model
        self.resume = resume
        self.timeline = timeline
        self.timeline_len = timeline.size
        self.x_shift = int(timeline[0])
        self.number_of_trajectories = number_of_trajectories
        self.clear_output = live_output_options['clear_output']
        species_mappings = model._listOfSpecies
        self.species = list(species_mappings.keys())
        self.number_species = len(self.species)
        self.current_trajectory = 1
        self.header_printed = False

    def trajectory_header(self):
        '''
        Create the trajectory header for the output.
        '''
        return "Trajectory (" + str(self.current_trajectory) + "/" + str(self.number_of_trajectories) + ")"

    def increment_trajectory(self, trajectory_num):
        '''
        Increment the trejectory counter.
        '''
        self.current_trajectory = trajectory_num + 1
        self.header_printed = False

    def print_text_header(self, file_obj):
        '''
        Print the header for text display type.

        :param file_obj: File object to write text output.
        :type file_obj: file object
        '''
        self.header_printed = True
        if self.number_of_trajectories > 1:
            print(self.trajectory_header(), file=file_obj)

        print("Time      |", end="", file=file_obj)
        for species in self.model.listOfSpecies:
            print(species[:10].ljust(10), end="|", file=file_obj)
        print("", file=file_obj)

    def display(self, curr_state, curr_time, trajectory_base, finished=False):
        '''
        Display the output for the live grapher.

        :param curr_state: Current state of the simulation. Should be a list of len 1 to get reference.
        :type curr_state: list

        :param curr_time: Current time of the simulation. Should be a list of len 1 to get reference.
        :type curr_time: list

        :param trajectory_base: Current results of the simulation.
        :type trajectory_base: list

        :param finished: Indicates whether or not the simulation has finished.
        :type finished: bool
        '''
        from IPython.display import clear_output # pylint: disable=import-outside-toplevel

        curr_time = curr_time[0]
        curr_state = curr_state[0]

        # necessary for __f function in hybrid solver
        if 't' in curr_state:
            if curr_state['t'] > curr_time:
                curr_time = curr_state['t']
        elif 'time' in curr_state:
            if curr_state['time'] > curr_time:
                curr_time = curr_state['time']

        if self.file_path is None or self.display_type == "graph":
            if self.clear_output:
                clear_output(wait=True)
            file_obj = sys.stdout
        else:
            mode = "w" if self.clear_output else "a"
            file_obj = open(self.file_path, mode, encoding="utf-8")

        if self.display_type == "text":

            if not self.header_printed:
                self.print_text_header(file_obj)

            print(str(round(curr_time, 2))[:10].ljust(10), end="|", file=file_obj)

            for i in range(self.number_species):
                print(str(curr_state[self.species[i]])[:10].ljust(10), end="|", file=file_obj)
            print("", file=file_obj)

        elif finished and self.display_type == "progress":
            print("progress = 100 %", file=file_obj)

        elif self.display_type == "progress":
            if self.number_of_trajectories > 1:
                print(self.trajectory_header(), file=file_obj)
            if self.resume is True:
                print(f"progress = {round(((curr_time-self.x_shift)/self.timeline_len)*100, 2)} %\n", file=file_obj)
            else:
                print(
                    f"progress = {round((curr_time / (self.timeline_len + self.x_shift)) * 100, 2) }%\n", file=file_obj
                )

        elif self.display_type == "graph":
            if finished:
                return

            import matplotlib.pyplot as plt # pylint: disable=import-outside-toplevel

            entry_count = floor(curr_time) - self.x_shift

            plt.figure(figsize=(18, 10))
            plt.xlim(right=self.timeline[-1])
            plt.xlim(left=self.timeline[0])
            plt.title(self.trajectory_header())
            for i in range(self.number_species):
                line_color = common_rgb_values()[(i) % len(common_rgb_values())]
                plt.plot(trajectory_base[0][:, 0][:entry_count].tolist(),
                         trajectory_base[0][:, i + 1][:entry_count].tolist(), color=line_color,
                         label=self.species[i])

                plt.plot([entry_count - 1, curr_time - self.timeline[0]], [trajectory_base[0][:, i + 1][entry_count - 1]
                    , curr_state[self.species[i]]], linewidth=3,
                         color=line_color)
            plt.legend(loc='upper right')
            plt.show()

        elif self.display_type == "figure":
            import plotly # pylint: disable=import-outside-toplevel
            import plotly.graph_objs as go # pylint: disable=import-outside-toplevel

            entry_count = floor(curr_time) - self.x_shift

            trace_list = []
            for i, species in enumerate(self.species):
                line_dict = {"color": common_rgb_values()[(i) % len(common_rgb_values())]}
                trace_list.append(
                    go.Scatter(
                        x=trajectory_base[0][:, 0][:entry_count].tolist(),
                        y=trajectory_base[0][:, i + 1][:entry_count].tolist(),
                        mode="lines", name=species, line=line_dict, legendgroup=species
                    )
                )

            layout = go.Layout(
                showlegend=True, title=self.trajectory_header(),
                xaxis={"range": [self.timeline[0], self.timeline[-1]]}
            )

            fig = dict(data=trace_list, layout=layout)
            json.dump(fig, file_obj, cls=plotly.utils.PlotlyJSONEncoder)

        if self.file_path is not None and self.display_type != "graph":
            file_obj.close()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import gillespy2
from example_models import create_telegraph_model
from gillespy2.core import Model, Species, Reaction, Parameter
from gillespy2.core.gillespyError import *
import os
import numpy as np


class TestTelegraphModel(unittest.TestCase):

    def test_telegraph_model(self):
        model = create_telegraph_model()
        solvers = [
            gillespy2.solvers.CLESolver,
            gillespy2.solvers.NumPySSASolver,
            gillespy2.solvers.ODECSolver,
            gillespy2.solvers.ODESolver,
            gillespy2.solvers.SSACSolver,
            gillespy2.solvers.TauHybridCSolver,
            gillespy2.solvers.TauHybridSolver,
            gillespy2.solvers.TauLeapingCSolver,
            gillespy2.solvers.TauLeapingSolver,
        ]
        for solver in solvers:
            with self.subTest(solver=solver.name):
                results = model.run(solver=solver)
                a = np.sum(results['ON'])/len(results['ON'])
                b = np.sum(results['OFF'])/len(results['OFF'])
                self.assertTrue( a < b) 



if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)
log = logging.getLogger()
log.setLevel(logging.WARN)
log.addHandler(_handler)

import numpy as np
from .ssa_solver import NumPySSASolver
from .ode_solver import ODESolver
from .tau_leaping_solver import TauLeapingSolver
from .tau_hybrid_solver import TauHybridSolver
from .CLE_solver import CLESolver
log.debug("Successful Import of NumPy solvers.")

__all__ = ['NumPySSASolver', 'ODESolver', 'TauLeapingSolver', 'TauHybridSolver', 'CLESolver']



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gillespy2.core.jsonify import Jsonify
from gillespy2.core.sortableobject import SortableObject

from gillespy2.core.gillespyError import ParameterError

class Parameter(SortableObject, Jsonify):
    """
    A parameter can be given as an expression (function) or directly
    as a value (scalar). If given an expression, it should be
    understood as evaluable in the namespace of a parent Model.

    :param name: The name by which this parameter is called or referenced in \
                 reactions, rate rules, assignment rules, and events.
    :type name: str

    :param expression: String for a function calculating parameter values. Should be
                       evaluable in namespace of Model.
    :type expression: str

    :raises ParameterError: Arg is of invalid type.  Required arg set to None.  Arg value is outside of accepted bounds.
    """

    def __init__(self, name=None, expression=None):
        # We allow expression to be passed in as a non-string type. Invalid strings
        # will be caught below. It is perfectly fine to give a scalar value as the expression.
        # This can then be evaluated in an empty namespace to the scalar value.
        self.value = None
        self.name = name

        if isinstance(expression, (int, float)):
            expression = str(expression)
        self.expression = expression

        self.validate()

    def __str__(self):
        return f"{self.name}: {self.expression}"

    def _evaluate(self, namespace=None):
        """
        Evaluate the expression and return the (scalar) value in the given
        namespace.

        :param namespace: The namespace in which to test evaulation of the parameter,
            if it involves other parameters, etc.
        :type namespace: dict

        :raises ParameterError: expression is of invalid type.  expression is set to None. \
                                expression is not evaluable within the given namespace.
        """
        if isinstance(self.expression, (int, float)):
            self.expression = str(self.expression)

        self.validate(coverage="expression")

        try:
            if namespace is None:
                namespace = {}
            self.value = float(eval(self.expression, namespace))
        except Exception as err:
            raise ParameterError(
                f"Could not evaluate expression: '{self.expression}'. Reason given: {err}."
            ) from err

    def sanitized_expression(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_expression = self.expression
        for i, name in enumerate(names):
            sanitized_expression = sanitized_expression.replace(
                name, "{"+str(i)+"}")
        return sanitized_expression.format(*replacements)

    def validate(self, expression=None, coverage="all"):
        """
        Validate the parameter.

        :param expression: String for a function calculating parameter values. Should be
                           evaluable in namespace of Model.
        :type expression: str

        :param coverage: The scope of attributes to validate.  Set to an attribute name to restrict validation \
                         to a specific attribute.
        :type coverage: str

        :raises ParameterError: Attribute is of invalid type.  Required attribute set to None.  \
                                Attribute value is outside of accepted bounds.
        """
        # Check name
        if coverage in ("all", "name"):
            if self.name is None:
                raise ParameterError("name can't be None type.")
            if not isinstance(self.name, str):
                raise ParameterError("name must be of type str.")
            if self.name == "":
                raise ParameterError("name can't be an empty string.")

        # Check expression
        if coverage in ("all", "expression"):
            if expression is None:
                expression = self.expression

            if expression is None:
                raise ParameterError("initial_value can't be None type.")
            if not isinstance(expression, str):
                raise ParameterError("expression must be of type str.")
            if expression == "":
                raise ParameterError("expression can't be an empty string.")



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np

import gillespy2
from example_models import create_decay, create_oregonator, create_michaelis_menten
from gillespy2.core.results import Results, Trajectory
from gillespy2 import SSACSolver
from gillespy2 import ODESolver
from gillespy2 import NumPySSASolver
from gillespy2 import TauLeapingSolver
from gillespy2 import TauHybridSolver
from gillespy2 import ODECSolver
from gillespy2 import TauLeapingCSolver
from gillespy2 import TauHybridCSolver


class TestAllSolvers(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.solvers = [
            SSACSolver,
            ODESolver,
            NumPySSASolver,
            TauLeapingSolver,
            TauHybridSolver,
            ODECSolver,
            TauLeapingCSolver,
            TauHybridCSolver,
        ]

        cls.sbml_features = {
            "AssignmentRule": lambda model, variable:
                model.add_assignment_rule(gillespy2.AssignmentRule(variable=variable, formula="1/(t+1)")),
            "RateRule": lambda model, variable:
                model.add_rate_rule(gillespy2.RateRule(variable=variable, formula="2*t")),
            "Event": lambda model, variable:
                model.add_event(gillespy2.Event(
                    trigger=gillespy2.EventTrigger(expression="t>1"),
                    assignments=[gillespy2.EventAssignment(variable=variable, expression="100")]
                )),
            "FunctionDefinition": lambda model, variable:
                model.add_function_definition(
                    gillespy2.FunctionDefinition(name="fn", function="variable", args=["variable"])),
        }

        # List of supported SBML features for each solver.
        # When a feature is implemented for a particular solver, add the feature to its list.
        cls.solver_supported_sbml_features = {
            NumPySSASolver: [],
            TauLeapingSolver: [],
            ODESolver: [],
            TauHybridSolver: [
                "AssignmentRule",
                "RateRule",
                "Event",
                "FunctionDefinition",
            ],

            SSACSolver: [],
            ODECSolver: [],
            TauLeapingCSolver: [],
            TauHybridCSolver: [
                "RateRule",
                "Event",
            ],
        }

        cls.model = create_decay()

        cls.results = {}
        cls.labeled_results = {}
        cls.labeled_results_more_trajectories = {}
    
    def test_extraneous_args(self):
        for solver in self.solvers:
            with self.subTest(solver=solver.name), self.assertLogs('GillesPy2', level='WARN'):
                model = create_decay()
                solver = solver(model=model)
                model.run(solver=solver, nonsense='ABC')

    def test_instantiated(self):
        for sp in self.model.listOfSpecies.values():
            sp.mode = 'discrete'

        for solver in self.solvers:
            with self.subTest(solver=solver.name):
                solver = solver(model=self.model)
                if "ODE" in solver.name:
                    self.labeled_results[solver.name] = self.model.run(solver=solver, number_of_trajectories=1)
                else:
                    self.labeled_results[solver.name] = self.model.run(solver=solver, number_of_trajectories=1, seed=1)
                self.labeled_results_more_trajectories[solver.name] = self.model.run(solver=solver, number_of_trajectories=2)

    def test_random_seed(self):
        for solver in self.solvers:
            with self.subTest(solver=solver.name):
                solver = solver(model=self.model)
                if "ODE" in solver.name:
                    same_results = self.model.run(solver=solver)
                    compare_results = self.model.run(solver=solver)
                else:
                    same_results = self.model.run(solver=solver, seed=1)
                    compare_results = self.model.run(solver=solver,seed=1)
                self.assertTrue(np.array_equal(same_results.to_array(), compare_results.to_array()))
                if solver.name in ["ODESolver", "ODECSolver"]: continue
                diff_results = self.model.run(solver=solver, seed=2)
                self.assertFalse(np.array_equal(diff_results.to_array(), same_results.to_array()))
    
    def test_random_seed_unnamed_reactions(self):
        model = self.model
        k2 = gillespy2.Parameter(name='k2', expression=1.0)
        model.add_parameter([k2])
        unnamed_rxn = gillespy2.Reaction(reactants={}, products={'Sp':1}, rate=k2)
        model.add_reaction(unnamed_rxn)
        for solver in self.solvers:
            with self.subTest(solver=solver.name):
                solver = solver(model=self.model)
                if "ODE" in solver.name:
                    same_results = self.model.run(solver=solver)
                    compare_results = self.model.run(solver=solver)
                else:
                    same_results = self.model.run(solver=solver, seed=1)
                    compare_results = self.model.run(solver=solver,seed=1)
                self.assertTrue(np.array_equal(same_results.to_array(), compare_results.to_array()))
                if solver.name in ["ODESolver", "ODECSolver"]: continue
                diff_results = self.model.run(solver=solver, seed=2)
                self.assertFalse(np.array_equal(diff_results.to_array(), same_results.to_array()))

    def test_return_type_show_labels(self):
        for solver in self.solvers:
            self.assertTrue(isinstance(self.labeled_results[solver.name], Results))
            self.assertTrue(isinstance(self.labeled_results[solver.name]['Sp'], np.ndarray))
            self.assertTrue(isinstance(self.labeled_results[solver.name]['Sp'][0], float))

            self.assertTrue(isinstance(self.labeled_results[solver.name][0], Trajectory))

            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver.name], Results))
            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver.name][0], Trajectory))
            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver.name][0]['Sp'], np.ndarray))
            self.assertTrue(isinstance(self.labeled_results_more_trajectories[solver.name][0]['Sp'][0], float))

    def test_sbml_feature_validation(self):
        def create_test_model(parameter_values=None):
            model = gillespy2.Model(name="TestModel")
            model.add_species(gillespy2.Species(name="S", initial_value=0))
            model.timespan(np.linspace(0, 10, 11))
            return model

        all_features = set(self.sbml_features.keys())
        for solver in self.solvers:
            unsupported_features = all_features.difference(self.solver_supported_sbml_features.get(solver))
            with self.subTest(solver=solver.name):
                for sbml_feature_name in unsupported_features:
                    model = create_test_model()
                    with self.subTest("Unsupported model features raise an error", sbml_feature=sbml_feature_name):
                        add_sbml_feature = self.sbml_features.get(sbml_feature_name)
                        add_sbml_feature(model, "S")
                        with self.assertRaises(gillespy2.ModelError):
                            solver.validate_sbml_features(model=model)

                for sbml_feature_name in self.solver_supported_sbml_features.get(solver):
                    model = create_test_model()
                    with self.subTest("Supported model features validate successfully", sbml_feature=sbml_feature_name):
                        add_sbml_feature = self.sbml_features.get(sbml_feature_name)
                        add_sbml_feature(model, "S")
                        solver.validate_sbml_features(model=model)

    def test_timeout(self):
        for solver in self.solvers:
            with self.subTest(solver=solver.name), self.assertLogs('GillesPy2', level='WARN'):
                model = create_oregonator()
                model.timespan(np.linspace(0, 1000000, 1001))
                solver = solver(model=model)
                model.run(solver=solver, timeout=0.1)

    def test_to_array(self):
        for solver in self.solvers:
            self.assertTrue(isinstance(self.labeled_results[solver.name].to_array()[0], np.ndarray))


if __name__ == '__main__':
    unittest.main()



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Class and methods for the CLE Solver"""
import copy
import random
import math
from threading import Thread, Event
import numpy as np
from gillespy2.solvers.utilities import Tau
from gillespy2.solvers.utilities import solverutils as nputils
from gillespy2.core import GillesPySolver, log, liveGraphing, SimulationError
from gillespy2.core import ModelError, ExecutionError
from gillespy2.core.results import Results


class CLESolver(GillesPySolver):
    """
    A Chemical Langevin Equation Solver for GillesPy2 models.

    This solver uses an algorithm that calculates multiple reactions in a single step over a given
    tau step size.  The change in propensities over this step are bounded by bounding the relative
    change in state, yielding greatly improved run-time performance with very little trade-off in accuracy.

    :param model: The model on which the solver will operate.
    :type model: gillespy2.Model
    """

    name = "CLESolver"
    rc = 0
    stop_event = None
    pause_event = None
    result = None

    def __init__(self, model=None, debug=False):
        if model is None:
            raise SimulationError("A model is required to run the simulation.")

        name = "CLESolver"
        rc = 0
        stop_event = None
        pause_event = None
        result = None
        self.model = model
        self.debug = debug
        self.is_instantiated = True

    def __get_reactions(self, step, curr_state, curr_time, save_time, propensities, reactions):
        r"""
        Helper Function to get reactions fired from t to t+tau.  
        
        In the equation:
        X_i(t+\tau) = X_i(t) + \sum_{j=1}^M \nu_{ji} \left[ a_j(x_t)\tau + \left( a_j(x_t)\tau \right)^{1/2} \mathcal{N}(0,1) \right]
        
        This function calculates: \left[ a_j(x_t)\tau + \left( a_j(x_t)\tau \right)^{1/2} \mathcal{N}(0,1) \right]
        for each reaction channel.
        Reference: Gillespie 2000. https://doi.org/10.1063/1.481811 (eq 21)

        :returns: Three values:
            rxn_count - dict with key=Reaction channel value=number of times fired
            curr_state - dict containing all state variables for system at current time
            curr_time - float representing current time
        """

        if curr_time + step > save_time:
            if self.debug:
                print("Step exceeds save_time, changing step size from ", step,
                      " to ", save_time - curr_time)
            step = save_time - curr_time

        if self.debug:
            print("Curr Time: ", curr_time, " Save time: ", save_time, "step: ", step)

        rxn_count = {}

        for rxn in reactions:
            #rxn_count[rxn] = np.random.poisson(propensities[rxn] * step)
            rxn_count[rxn] = propensities[rxn]*step + \
                math.sqrt(propensities[rxn]*step)*np.random.normal(0,1)

        if self.debug:
            print("Reactions Fired: ", rxn_count)

        curr_time = curr_time+step

        return rxn_count, curr_state, curr_time

    @classmethod
    def get_solver_settings(cls):
        """
        Returns a list of arguments supported by CLE_solver.run.
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        :rtype: tuple
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'seed', 'debug', 'profile','timeout', 'tau_tol')

    def run(self=None, model=None, t=None, number_of_trajectories=1, increment=None, seed=None,
            debug=False, profile=False,  live_output=None, live_output_options={},
            timeout=None, resume=None, tau_tol=0.03, **kwargs):
        """
        Function calling simulation of the model.
        This is typically called by the run function in GillesPy2 model objects
        and will inherit those parameters which are passed with the model
        as the arguments this run function.

        :param model: The model on which the solver will operate. (Deprecated)
        :type model: gillespy2.Model

        :param t: Simulation run time.
        :type t: int or float

        :param number_of_trajectories: The number of times to sample the chemical master equation. Each
        trajectory will be returned at the end of the simulation. By default number_of_trajectories = 1.
        :type number_of_trajectories: int

        :param increment: Save point increment for recording data.
        :type increment: float

        :param seed: The random seed for the simulation. Optional, defaults to None.
        :type seed: int

        :param debug: Set to True to provide additional debug information about the simulation.
        :type debug: bool

        :param profile: Set to True to provide information about step size (tau) taken at each step.
        :type profile: bool

        :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :type live_output: str

        :param live_output_options: Contains options for live_output. By default {"interval":1}. "interval"
            specifies seconds between displaying. "clear_output" specifies if display should be refreshed with each
            display.
        :type live_output_options: dict

        :param timeout: If set, if simulation takes longer than timeout, will exit.
        :type timeout: int
        
        :param resume: Result of a previously run simulation, to be resumed.
        :type resume: gillespy2.Results
        
        :param tau_tol: Tolerance level for Tau leaping algorithm.  Larger tolerance values will
            result in larger tau steps. Default value is 0.03.
        :type tau_tol: float

        :returns: A result object containing the results of the simulation.
        :rtype: gillespy2.Results
        """
        from gillespy2 import log

        if self is None:
            # Post deprecation block
            # raise SimulationError("CLESolver must be instantiated to run the simulation")
            # Pre deprecation block
            log.warning(
                """
                `gillespy2.Model.run(solver=CLESolver)` is deprecated.

                You should use `gillespy2.Model.run(solver=CLESolver(model=gillespy2.Model))
                Future releases of GillesPy2 may not support this feature.
                """
            )
            self = CLESolver(model=model, debug=debug, profile=profile)
        
        if model is not None:
            log.warning('model = gillespy2.model is deprecated. Future releases '
                        'of GillesPy2 may not support this feature.')
        if self.model is None:
            if model is None:
                raise SimulationError("A model is required to run the simulation.")
            self.model = copy.deepcopy(model)
        
        self.model.compile_prep()
        self.validate_model(self.model, model)
        self.validate_sbml_features(model=self.model)

        self.validate_tspan(increment=increment, t=t)
        if increment is None:
            increment = self.model.tspan[-1] - self.model.tspan[-2]
        if t is None:
            t = self.model.tspan[-1]

        self.stop_event = Event()
        self.pause_event = Event()

        if timeout is not None and timeout <= 0:
            timeout = None
        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

        # create numpy array for timeline
        if resume is not None:
            # start where we last left off if resuming a simulatio
            lastT = resume['time'][-1]
            step = lastT - resume['time'][-2]
            timeline = np.arange(lastT, t+step, step)
        else:
            timeline = np.linspace(0, t, int(round(t / increment + 1)))

        species = list(self.model._listOfSpecies.keys())
        trajectory_base, tmpSpecies = nputils.numpy_trajectory_base_initialization(self.model, number_of_trajectories,
                                                                                   timeline, species, resume=resume)

        # total_time and curr_state are list of len 1 so that __run receives reference
        if resume is not None:
            total_time = [resume['time'][-1]]
        else:
            total_time = [0]

        curr_state = [None]
        live_grapher = [None]

        sim_thread = Thread(target=self.___run, args=(curr_state, total_time, timeline, trajectory_base, tmpSpecies,
                                                      live_grapher,), kwargs={'t': t,
                                                                              'number_of_trajectories':
                                                                                  number_of_trajectories,
                                                                              'increment': increment, 'seed': seed,
                                                                              'debug': debug, 'resume': resume,
                                                                              'timeout': timeout, 'tau_tol': tau_tol
                                                                              })
        try:
            time = 0
            sim_thread.start()
            if live_output is not None:
                import gillespy2.core.liveGraphing
                live_output_options['type'] = live_output
                gillespy2.core.liveGraphing.valid_graph_params(
                    live_output_options)
                if resume is not None:
                    resumeTest = True  # If resuming, relay this information to live_grapher
                else:
                    resumeTest = False
                live_grapher[
                    0] = gillespy2.core.liveGraphing.LiveDisplayer(self.model,
                                                                   timeline,
                                                                   number_of_trajectories,
                                                                   live_output_options,
                                                                   resume=resumeTest)
                display_timer = gillespy2.core.liveGraphing.RepeatTimer(
                    live_output_options['interval'],
                    live_grapher[0].display, args=(curr_state,
                                                   total_time,
                                                   trajectory_base,
                                                   live_output
                                                   )
                    )
                display_timer.start()

            if timeout is not None:
                while sim_thread.is_alive():
                    sim_thread.join(.1)
                    time += .1
                    if time >= timeout:
                        break
            else:
                while sim_thread.is_alive():
                    sim_thread.join(.1)

            if live_grapher[0] is not None:
                display_timer.cancel()
            self.stop_event.set()
            while self.result is None:
                pass
        except KeyboardInterrupt:
            if live_output:
                display_timer.pause = True
                display_timer.cancel()
            self.pause_event.set()
            while self.result is None:
                pass
        if hasattr(self, 'has_raised_exception'):
            raise SimulationError(
                f"Error encountered while running simulation:\nReturn code: {int(self.rc)}.\n"
            ) from self.has_raised_exception

        return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, curr_state,total_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, seed=None, debug=False, profile=False,
               timeout=None, resume=None, tau_tol=0.03, **kwargs):

        try:
            self.__run(curr_state, total_time, timeline, trajectory_base, tmpSpecies, live_grapher, t, number_of_trajectories,
                       increment, seed, debug, profile, timeout, resume, tau_tol, **kwargs)

        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, curr_state, total_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, seed=None, debug=False, profile=False, timeout=None,
              resume=None, tau_tol=0.03, **kwargs):

        # for use with resume, determines how much excess data to cut off due to
        # how species and time are initialized to 0
        timeStopped = 0
        if resume is not None:
            if resume[0].model != self.model:
                raise ModelError('When resuming, one must not alter the model being resumed.')
            if t < resume['time'][-1]:
                raise ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")
        if debug:
            print("t = ", t)
            print("increment = ", increment)

        species_mappings, species, parameter_mappings, number_species = nputils.numpy_initialization(self.model)

        if seed is not None:
            if not isinstance(seed, int):
                seed = int(seed)
            if seed > 0:
                random.seed(seed)
                np.random.seed(seed)
            else:
                raise ModelError('seed must be a positive integer')

        simulation_data = []

        for trajectory_num in range(number_of_trajectories):
            if self.stop_event.is_set():
                self.rc = 33
                break
            elif self.pause_event.is_set():
                timeStopped = timeline[entry_count]
            # For multi trajectories, live_grapher needs to be informed of trajectory increment
            if live_grapher[0] is not None:
                live_grapher[0].increment_trajectory(trajectory_num)

            start_state = [0] * (len(self.model.listOfReactions) + len(self.model.listOfRateRules))
            propensities = {}
            curr_state[0] = {}

            if resume is not None:
                save_time = total_time[0]
                curr_time = [save_time]
            else:
                save_time = 0
                curr_time = [0]

            curr_state[0]['vol'] = self.model.volume
            data = {'time': timeline}
            steps_taken = []
            steps_rejected = 0
            entry_count = 0
            trajectory = trajectory_base[trajectory_num]

            HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, critical_threshold = Tau.initialize(self.model, tau_tol)

            if resume is not None:
                for spec in self.model.listOfSpecies:
                    curr_state[0][spec] = tmpSpecies[spec]
            else:
                for spec in self.model.listOfSpecies:
                    curr_state[0][spec] = self.model.listOfSpecies[spec].initial_value

            for param in self.model.listOfParameters:
                curr_state[0][param] = self.model.listOfParameters[param].value

            for i, rxn in enumerate(self.model.listOfReactions):
                # set reactions to uniform random number and add to start_state
                start_state[i] = (math.log(random.uniform(0, 1)))
                if debug:
                    print("Setting Random number ",
                          start_state[i], " for ", self.model.listOfReactions[rxn].name)

            compiled_propensities = {}
            for i, r in enumerate(self.model.listOfReactions):
                compiled_propensities[r] = compile(self.model.listOfReactions[r].ode_propensity_function, '<string>', 'eval')

            timestep = 0
            
            # Each save step
            while entry_count < timeline.size:
                if self.stop_event.is_set():
                    self.rc = 33
                    break
                elif self.pause_event.is_set():
                    timeStopped = timeline[entry_count]
                    break

                # Until save step reached
                while curr_time[0] < save_time:
                    if self.stop_event.is_set():
                        self.rc = 33
                        break
                    elif self.pause_event.is_set():
                        timeStopped = timeline[entry_count]
                        break

                    propensity_sum = 0

                    for i, r in enumerate(self.model.listOfReactions):
                        propensities[r] = eval(compiled_propensities[r], curr_state[0])
                        propensity_sum += propensities[r]

                    tau_args = [HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, tau_tol, critical_threshold,
                                self.model, propensities, curr_state[0], curr_time[0], save_time]

                    tau_step = Tau.select(*tau_args)

                    prev_start_state = start_state.copy()
                    prev_curr_state = curr_state[0].copy()
                    prev_curr_time = curr_time[0]
                    total_time[0] = curr_time[0]

                    loop_cnt = 0
                    while True:
                        loop_cnt += 1
                        if loop_cnt > 100:
                            raise Exception("Loop over __get_reactions() exceeded loop count")

                        reactions, curr_state[0], curr_time[0] = self.__get_reactions(
                            tau_step, curr_state[0], curr_time[0], save_time,
                            propensities, self.model.listOfReactions)

                        # Update curr_state with the result of the CLE step
                        species_modified = {}
                        for i, rxn in enumerate(self.model.listOfReactions):
                            if reactions[rxn] > 0:
                                for reactant in self.model.listOfReactions[rxn].reactants:
                                    species_modified[reactant.name] = True
                                    curr_state[0][reactant.name] -= self.model.listOfReactions[
                                        rxn].reactants[reactant] * reactions[rxn]
                                for product in self.model.listOfReactions[rxn].products:
                                    species_modified[product.name] = True
                                    curr_state[0][product.name] += self.model.listOfReactions[
                                        rxn].products[product] * reactions[rxn]
                        neg_state = False
                        for spec in species_modified:
                            if curr_state[0][spec] < 0:
                                neg_state = True
                                if debug:
                                    print("Negative state detected: curr_state[{0}]= {1}".format(spec,
                                                                                                 curr_state[0][spec]))
                        if neg_state:
                            if debug:
                                print("\trxn={0}".format(reactions))
                            start_state = prev_start_state.copy()
                            curr_state[0] = prev_curr_state.copy()
                            curr_time[0] = prev_curr_time
                            total_time[0] = prev_curr_time

                            tau_step = tau_step / 2
                            steps_rejected += 1
                            if debug:
                                print("Resetting curr_state[{0}]= {1}".format(spec, curr_state[0][spec]))
                            if debug:
                                print("\tRejecting step, taking step of half size, tau_step={0}".format(tau_step))
                        else:
                            break  # breakout of the while True

                # save step reached
                for i in range(number_species):
                    trajectory[entry_count][i + 1] = curr_state[0][species[i]]
                save_time += increment
                timestep += 1
                entry_count += 1
                
            # end of trajectory
            for i in range(number_species):
                data[species[i]] = trajectory[:, i+1]
            simulation_data.append(data)

            if profile:
                print(steps_taken)
                print("Total Steps Taken: ", len(steps_taken))
                print("Total Steps Rejected: ", steps_rejected)

        # If simulation has been paused, or tstopped !=0
        if timeStopped != 0 or resume is not None:
            simulation_data = nputils.numpy_resume(timeStopped, simulation_data, resume=resume)

        self.result = simulation_data
        return self.result, self.rc
